from matplotlib import pyplot as plt
from matplotlib.colors import LinearSegmentedColormap
import networkx as nx
import numpy as np
import pandas as pd
import matplotlib.gridspec as gridspec

from . import plot_helper, an
from .CCIData_class import CCIData


def network_plot(
    network,
    p_vals=None,
    diff_plot=False,
    normalise=True,
    remove_unconnected=True,
    show_labels=False,
    p_val_cutoff=0.05,
    edge_weight=50,
    text_size=15,
    node_size=2500,
    figsize=None,
    arrowsize=20,
    node_label_dist=1,
    p_val_text_size=10,
    node_colors=None,
    node_palette="tab20",
    outer_node_palette=None,
    show=True,
    show_legend=True,
    legend_size=12,
    title=None,
    title_size=14
):
    """Plots a network with optional edge significance highlighting and node coloring based on in-degree and out-degree difference.

    Args:
        network (pandas.DataFrame or numpy.ndarray): The adjacency matrix representing the network.
        p_vals (pandas.DataFrame or numpy.ndarray, optional): A matrix of p-values corresponding to the edges in `network`. If not provided, significance values will not be plotted. Defaults to None.
        diff_plot (bool, optional): Whether you are plotting the network difference, to show up and down-regulated edges. Defaults to False.
        normalise (bool, optional): Whether to normalize the network matrix before plotting. Defaults to True.
        remove_unconnected (bool, optional): Whether to remove cell types that do not interact with any cell types. Defaults to True.
        show_labels (bool, optional): Whether to show node labels. Defaults to True.
        p_val_cutoff (float, optional): The p-value cutoff for determining significant edges. Defaults to 0.05.
        edge_weight (float, optional): The base weight for edges. Defaults to 20.
        text_size (int, optional): The font size for node labels. Defaults to 15.
        node_size (int, optional): The size of the nodes. Defaults to 2500.
        figsize (tuple, optional): The size of the figure. Defaults to None.
        arrowsize (int, optional): The size of the arrow heads for edges. Defaults to 50.
        node_label_dist (float, optional): A factor for adjusting the distance between nodes and labels. Defaults to 1.
        p_val_text_size (int, optional): The font size for p-value labels. Defaults to 10.
        node_colors (dict, optional): A dictionary of colors for each node. Overwrites node_palette. Defaults to None.
        node_palette (str, optional): The name of the color palette to use for nodes. Defaults to "tab20".
        outer_node_palette (str, optional): The name of the color palette to use for outer nodes to show sender/reciever nodes. Defaults to None.
        show (bool, optional): Whether to show the plot or not. Defaults to True.
        show_legend (bool, optional): Whether to show legend. Defaults to False.
        legend_size (int, optional): Font size for legend. Defaults to 12.
        title (str, optional): Title of the plot. Defaults to None.
        title_size (int, optional): Font size for title. Defaults to 14.

    Returns:
        tuple: A tuple containing the figure and axis objects.
    """

    if not isinstance(network, pd.DataFrame):
        raise ValueError("Input should be a dataframe.")
    
    if figsize is None:
        if show_legend:
            figsize = (10, 8)
        else:
            figsize = (8, 8)

    # Adjust the figure layout to accommodate the legend on the right
    fig = plt.figure(figsize=figsize)
    gs = gridspec.GridSpec(1, 2, width_ratios=[5, 1], wspace=0.2)

    # Main plot area
    ax = fig.add_subplot(gs[0])
    plt.sca(ax)  # Set the current axis to the main plot area

    if remove_unconnected:
        cell_types = (network != 0).any(axis=0) + (network != 0).any(axis=1)
        network = network.loc[cell_types, cell_types]

    if normalise:
        if network.min().min() >= 0:
            network = network / network.sum().sum()
        else:
            network = network / network.abs().sum().sum()
    network_abs = abs(network)

    if normalise:
        network_abs = network_abs / network_abs.sum().sum()
    network_abs = network.astype(float)
    G_network = nx.from_pandas_adjacency(network_abs, create_using=nx.DiGraph)
    pos = nx.circular_layout(G_network)
    weights = nx.get_edge_attributes(G_network, "weight")

    # Calculate the in-degree and out-degree for each node
    in_degree = dict(G_network.in_degree(weight="weight"))
    out_degree = dict(G_network.out_degree(weight="weight"))

    in_out_diff = {node: in_degree[node] - out_degree[node] for node in G_network.nodes}

    max_diff = max(abs(value) for value in in_out_diff.values())
    color_scale = np.linspace(-max_diff, max_diff, 256)

    if outer_node_palette is None:
        # Create a color scale based on the in-degree and out-degree difference
        cmap_colors = [(1, 0, 0), (0.7, 0.7, 0.7), (0, 0, 1)]  # Blue, Grey, Red
        outer_node_cmap = LinearSegmentedColormap.from_list("custom_cmap", cmap_colors)
    else:
        outer_node_cmap = plt.get_cmap(outer_node_palette)

    
    edge_colors = []
    # Map node colors to the in-degree and out-degree difference
    if sum(abs(value) for value in in_out_diff.values()) == 0:
        edge_colors = ['grey' for node in G_network.nodes]
    else:
        edge_colors = [
            outer_node_cmap(int(np.interp(in_out_diff[node], color_scale, range(256))))
            for node in G_network.nodes
        ]

    if node_colors is not None:
        node_colors_list = [node_colors[node] for node in G_network.nodes]
    else:
        if node_palette is not None:
            node_colors_list = \
                list(plt.get_cmap(node_palette).colors)[:len(G_network.nodes)]
        else:
            node_colors_list = ["grey" for node in G_network.nodes]

    if p_vals is None or diff_plot == False:
        # Create a non-significant matrix
        p_vals = network_abs.replace(network_abs.values, 1, inplace=False)
    else:
        # Prevent removal of pvals of 0
        p_vals[p_vals == 0] = 1e-300

    # Get edges that are significant
    G_p_vals = nx.from_pandas_adjacency(p_vals, create_using=nx.DiGraph)
    G_network_updown = nx.from_pandas_adjacency(network, create_using=nx.DiGraph)

    non_sig_up = [
        (u, v) for (u, v, d) in G_p_vals.edges(data=True)
        if d["weight"] > p_val_cutoff
        and u in G_network_updown
        and v in G_network_updown[u]
        and G_network_updown[u][v]["weight"] > 0]
    non_sig_up = [edge for edge in non_sig_up if edge in weights.keys()]

    non_sig_down = [
        (u, v) for (u, v, d) in G_p_vals.edges(data=True)
        if d["weight"] > p_val_cutoff
        and u in G_network_updown
        and v in G_network_updown[u]
        and G_network_updown[u][v]["weight"] < 0]
    non_sig_down = [edge for edge in non_sig_down if edge in weights.keys()]

    sig_up = [
        (u, v) for (u, v, d) in G_p_vals.edges(data=True)
        if d["weight"] <= p_val_cutoff
        and u in G_network_updown
        and v in G_network_updown[u]
        and G_network_updown[u][v]["weight"] > 0]
    sig_up = [edge for edge in sig_up if edge in weights.keys()]

    sig_down = [
        (u, v) for (u, v, d) in G_p_vals.edges(data=True) if d["weight"] <= p_val_cutoff
        and u in G_network_updown
        and v in G_network_updown[u]
        and G_network_updown[u][v]["weight"] < 0]
    sig_down = [edge for edge in sig_down if edge in weights.keys()]

    edge_thickness_non_sig_up = []
    edge_thickness_non_sig_down = []
    edge_thickness_sig_up = []
    edge_thickness_sig_down = []

    for edge in weights.keys():
        if edge in non_sig_up:
            edge_thickness_non_sig_up.append(weights[edge] * edge_weight)
        else:
            edge_thickness_non_sig_up.append(0)

        if edge in non_sig_down:
            edge_thickness_non_sig_down.append(weights[edge] * edge_weight)
        else:
            edge_thickness_non_sig_down.append(0)

        if edge in sig_up:
            edge_thickness_sig_up.append(weights[edge] * edge_weight)
        else:
            edge_thickness_sig_up.append(0)

        if edge in sig_down:
            edge_thickness_sig_down.append(weights[edge] * edge_weight)
        else:
            edge_thickness_sig_down.append(0)

    # if node_colors is None:
    nx.draw_networkx_nodes(
        G_network,
        pos,
        node_size=node_size,
        node_color=node_colors_list,
        edgecolors=edge_colors,
        linewidths=8.0,
    )

    if diff_plot:
        # Draw non-self edges first
        nx.draw_networkx_edges(
            G_network,
            pos,
            node_size=node_size * 2,
            connectionstyle="arc3,rad=0.08",
            width=edge_thickness_non_sig_up,
            arrows=True,
            arrowstyle="->",
            arrowsize=arrowsize,
            edge_color="pink"
        )
        
        # Same pattern for non-sig down edges
        nx.draw_networkx_edges(
            G_network,
            pos,
            node_size=node_size * 2,
            connectionstyle="arc3,rad=0.08",
            width=edge_thickness_non_sig_down,
            arrows=True,
            arrowstyle="->",
            arrowsize=arrowsize,
            edge_color="lightgreen"
            )

    else:
        # Non-self edges
        nx.draw_networkx_edges(
            G_network,
            pos,
            node_size=node_size * 2,
            connectionstyle="arc3,rad=0.08",
            width=edge_thickness_non_sig_up,
            arrows=True,
            arrowstyle="->",
            arrowsize=arrowsize,
        )

        # Same for non-sig down edges
        nx.draw_networkx_edges(
            G_network,
            pos,
            node_size=node_size * 2,
            connectionstyle="arc3,rad=0.08",
            width=edge_thickness_non_sig_down,
            arrows=True,
            arrowstyle="->",
            arrowsize=arrowsize,
        )

    # Significant up edges
    nx.draw_networkx_edges(
        G_network,
        pos,
        node_size=node_size * 2,
        connectionstyle="arc3,rad=0.08",
        width=edge_thickness_sig_up,
        arrows=True,
        arrowstyle="->",
        arrowsize=arrowsize,
        edge_color="purple",
    )

    # Significant down edges  
    nx.draw_networkx_edges(
        G_network,
        pos,
        node_size=node_size * 2,
        connectionstyle="arc3,rad=0.08",
        width=edge_thickness_sig_down,
        arrows=True,
        arrowstyle="->",
        arrowsize=arrowsize,
        edge_color="green",
    )

    edge_labels = nx.get_edge_attributes(G_p_vals, "weight")
    edge_labels = {
        key: edge_labels[key] for key in G_network.edges().keys() if key in edge_labels
    }

    # Add edge labels for significant edges
    for key, value in edge_labels.items():
        if value > p_val_cutoff:
            edge_labels[key] = ""
        else:
            edge_labels[key] = round(value, 3)

    def offset(d, pos, dist=0.05, loop_shift=0.1):
        for (u, v), obj in d.items():
            if u != v:
                par = dist * (pos[v] - pos[u])
                dx, dy = par[1], -par[0]
                x, y = obj.get_position()
                obj.set_position((x + dx, y + dy))
            else:
                x, y = obj.get_position()
                obj.set_position((x, y + loop_shift))

    d = nx.draw_networkx_edge_labels(
        G_network, 
        pos, 
        edge_labels, 
        font_size=p_val_text_size, 
        connectionstyle="arc3,rad=0.08"
        )

    offset(d, pos)

    pos.update(
        (x, [y[0] * 1.4 * node_label_dist, y[1] * (1.25 + 0.05) * node_label_dist])
        for x, y in pos.items()
    )

    if show_labels:
        nx.draw_networkx_labels(
            G_network,
            pos,
            font_weight="bold",
            font_color="black",
            font_size=text_size,
            clip_on=False,
            horizontalalignment="center",
        )

    ax = plt.gca()
    ax.margins(0.08)
    plt.axis("off")
    
    if show_legend:
        # Create a color bar in the bottom-right corner
        color_bar_ax = fig.add_axes([0.77, 0.2, 0.05, 0.2])
        if sum(abs(value) for value in in_out_diff.values()) != 0:
            sm = plt.cm.ScalarMappable(cmap=outer_node_cmap,
                                    norm=plt.Normalize(vmin=-max_diff, vmax=max_diff))
            sm.set_array([])
            cbar = plt.colorbar(sm, cax=color_bar_ax)
            cbar.set_ticks([])  # Remove the ticks
            cbar.set_label('Net sender ← → Net receiver', fontsize=legend_size)

        # Add the legend in the top-right corner
        legend_elements = []
        if diff_plot:
            legend_elements.extend([
                plt.Line2D([0], [0], color='pink', lw=2, 
                           label='Non-significant positive'),
                plt.Line2D([0], [0], color='lightgreen', lw=2, 
                           label='Non-significant negative'),
                plt.Line2D([0], [0], color='purple', lw=2, 
                           label='Significant positive'),
                plt.Line2D([0], [0], color='green', lw=2, 
                           label='Significant negative')
            ])

        if node_colors is not None:
            for node, color in node_colors.items():
                if node in network.index:
                    legend_elements.append(
                        plt.Line2D([0], [0], marker='o', color='w',
                                markerfacecolor=color, markersize=10, label=node)
                    )
        else:
            for i in range(len(G_network.nodes)):
                legend_elements.append(
                    plt.Line2D([0], [0], marker='o', color='w',
                            markerfacecolor=node_colors_list[i], markersize=10, 
                            label=list(G_network.nodes.keys())[i])
                )

        if legend_elements:
            legend_ax = fig.add_axes([0.7, 0.55, 0.2, 0.3])
            legend_ax.axis('off')
            legend_ax.legend(handles=legend_elements, loc='center', 
                             fontsize=legend_size)

    if title is not None:
        plt.suptitle(title, fontsize=title_size, y = 0.875, fontweight="bold")

    plt.tight_layout()
    if show:
        plt.show()
    else:
        return fig, ax


def chord_plot(
    network,
    min_int=0.001,
    n_top_ccis=10,
    colors=None,
    show=True,
    title=None,
    title_size=14,
    label_size=10,
    figsize=None,
    show_legend=False,
    legend_size=12
):
    """Plots a chord plot of a network

    Args:
        network (pandas.DataFrame or numpy.ndarray): The adjacency matrix representing the network.
        min_int (float): Minimum interactions to display cell type. Defaults to 0.01.
        n_top_ccis (int): Number of top cell types to display. Defaults to 10.
        colors (dict): Dict of colors for each cell type to use for the plot. Defaults to None.
        show (bool): Whether to show plot or not. Defaults to True.
        title (str): Title of the plot. Defaults to None.
        title_size (int): Font size of the title. Defaults to 14.
        label_size (int): Font size of the labels. Defaults to None.
        figsize (tuple): Size of the figure. Defaults to None.
        show_legend (bool): Whether to show legend. Defaults to False.
        legend_size (int): Font size for legend. Defaults to 12.

    Returns:
        tuple: A tuple containing the figure and axis objects.
    """

    if not isinstance(network, pd.DataFrame):
        raise ValueError("Input should be a dataframe.")

    network = network.transpose()
    
    if figsize is None:
        if show_legend:
            figsize = (10, 8)
        else:
            figsize = (8, 8)
    
    # Create figure with gridspec to accommodate legend
    fig = plt.figure(figsize=figsize)
    if show_legend:
        gs = gridspec.GridSpec(1, 2, width_ratios=[4, 1], wspace=0)
        ax = fig.add_subplot(gs[0])
    else:
        ax = plt.axes([0, 0, 1, 1])

    flux = network.values

    total_ints = flux.sum(axis=1) + flux.sum(axis=0) - flux.diagonal()
    keep = total_ints > min_int
    # Limit of 10 for good display #
    if sum(keep) > n_top_ccis:
        keep = np.argsort(-total_ints)[0:n_top_ccis]
    flux = flux[:, keep]
    flux = flux[keep, :].astype(float)
    cell_names = network.index.values.astype(str)[keep]
    nodes = cell_names

    color_list = []
    if colors is not None:
        for cell in cell_names:
            color_list.append(colors[cell])
    else:
        color_list = None

    nodePos = plot_helper.chordDiagram(flux, ax, lim=1.25, colors=color_list)
    ax.axis("off")
    prop = dict(fontsize=label_size, ha="center", va="center")

    for i in range(len(cell_names)):
        x, y = nodePos[i][0:2]
        if label_size != 0:
            ax.text(x, y, nodes[i], rotation=nodePos[i][2], **prop)

    if show_legend and colors is not None:
        # Add the legend in the right subplot
        legend_ax = fig.add_subplot(gs[1])
        legend_ax.axis('off')
        
        legend_elements = []
        for node, color in colors.items():
            if node in cell_names:
                legend_elements.append(
                    plt.Line2D([0], [0], marker='o', color='w',
                            markerfacecolor=color, markersize=10, label=node)
                )
        
        if legend_elements:
            legend_ax.legend(handles=legend_elements, loc='center', 
                             fontsize=legend_size)

    if title is not None:
        fig.suptitle(title, fontsize=title_size, y=0.95, fontweight="bold")

    plt.tight_layout()

    if show:
        plt.show()
    else:
        return fig, ax


def dissim_hist(
    dissimilarity_scores,
    x_label_size=18,
    y_label_size=24,
    x_tick_size=14,
    y_tick_size=12,
    figsize=(6, 5),
    show=True,
    title=None,
    title_size=14
):
    """Plots a histogram of dissimilarity scores.

    Args:
        dissimilarity_scores (dict): A dictionary of dissimilarity scores.
        x_label_size (int): Font size for x-axis label. Defaults to 18.
        y_label_size (int): Font size for y-axis label. Defaults to 24.
        x_tick_size (int): Font size for ticks. Defaults to 14.
        y_tick_size (int): Font size for ticks. Defaults to 12.
        figsize (tuple): Size of the figure. Defaults to (10, 8).
        show (bool): Whether to show the plot or not. Defaults to True.
        title (str): Title of the plot. Defaults to None.
        title_size (int): Font size of the title. Defaults to 14.

    Returns:
        matplotlib.figure.Figure: The figure
    """

    fig = plt.figure(figsize=figsize)
    plt.style.use('default')
    plt.hist(list(dissimilarity_scores.values()))
    plt.xlim(0, 1)
    plt.xlabel("Dissimilarity Score", fontsize=x_label_size)
    plt.ylabel("Count", fontsize=y_label_size)
    plt.tick_params(axis='x', which='major', labelsize=x_tick_size)
    plt.tick_params(axis='y', which='major', labelsize=y_tick_size)

    if title is not None:
        plt.title(title, fontsize=title_size, pad=20)

    if show:
        plt.show()
    else:
        plt.close(fig)
        return fig


def lr_top_dissimilarity(
    dissimilarity_scores,
    n=10,
    top=True,
    x_label_size=18,
    y_label_size=24,
    x_tick_size=14,
    y_tick_size=12,
    figsize=(6, 5),
    show=True,
    title=None,
    title_size=14
):
    """Plots a bar plot of LR pairs with highest/lowest dissimilarity scores.

    Args:
        dissimilarity_scores (dict): A dictionary of dissimilarity scores.
        n (int): Number of LR pairs to plot.
        top (bool): If True, plot LR pairs with highest dissimilarity scores. If False, plot LR pairs with lowest dissimilarity scores.
        x_label_size (int): Font size for x-axis label. Defaults to 18.
        y_label_size (int): Font size for y-axis label. Defaults to 24.
        x_tick_size (int): Font size for ticks. Defaults to 14.
        y_tick_size (int): Font size for ticks. Defaults to 12.
        figsize (tuple): Size of the figure. Defaults to (10, 8).
        show (bool): Whether to show the plot or not. Defaults to True.
        title (str): Title of the plot. Defaults to None.
        title_size (int): Font size of the title. Defaults to 14.

    Returns:
        matplotlib.figure.Figure: The figure
    """

    reverse = not top
    sorted_items = sorted(
        dissimilarity_scores.items(), key=lambda x: x[1], reverse=reverse
    )
    top_n_items = sorted_items[-n:]
    keys, values = zip(*top_n_items)

    fig = plt.figure(figsize=figsize)
    plt.style.use('default')
    plt.barh(keys, values)
    plt.xlabel("Dissimilarity Score", fontsize=x_label_size)
    plt.ylabel("LR Pair", fontsize=y_label_size)
    plt.tick_params(axis='x', which='major', labelsize=x_tick_size)
    plt.tick_params(axis='y', which='major', labelsize=y_tick_size)
    
    if title is not None:
        plt.title(title, fontsize=title_size, pad=20)

    if show:
        plt.show()
    else:
        plt.close(fig)
        return fig
    

def lr_barplot(
    sample,
    assay="raw",
    n=15,
    x_label_size=18,
    y_label_size=24,
    x_tick_size=14,
    y_tick_size=12,
    figsize=(6, 5),
    show=True,
    title=None,
    title_size=14
):
    """Plots a bar plot of LR pairs and their proportions for a sample.

    Args:
        sample (CCIData): The CCIData object.
        assay (str): The assay to use. Defaults to "raw".
        n (int): Number of LR pairs to plot. If None, plot all LR pairs. Defaults to 15.
        x_label_size (int): Font size for x-axis label. Defaults to 18.
        y_label_size (int): Font size for y-axis label. Defaults to 24.
        x_tick_size (int): Font size for tick labels. Defaults to 14.
        y_tick_size (int): Font size for tick labels. Defaults to 12.
        figsize (tuple): Size of the figure. Defaults to (10, 8).
        title (str) (optional): Title for the plot. Defaults to None.
        title_size (int): Font size of the title. Defaults to 14.

    Returns:
        matplotlib.figure.Figure: The figure
    """

    if assay not in sample.assays:
        raise ValueError("Assay not found in sample.")

    interactions = [(lr, df.sum().sum()) for (lr, df) \
        in sample.assays[assay]['cci_scores'].items()]
    interactions.sort(key=lambda x: x[1])
    interactions = interactions[-n:]
    keys, values = zip(*interactions)
    values = [value / sum(values) for value in values]

    fig = plt.figure(figsize=figsize)
    plt.style.use('default')
    plt.barh(keys, values)
    plt.xlabel("Relative Interaction Strength", fontsize=x_label_size)
    plt.ylabel("LR Pair", fontsize=y_label_size)
    plt.tick_params(axis='x', which='major', labelsize=x_tick_size)
    plt.tick_params(axis='y', which='major', labelsize=y_tick_size)

    if title:
        plt.title(title, pad=20, fontsize=title_size)
    
    plt.tight_layout()

    if show:
        plt.show()
    else:
        plt.close(fig)
        return fig


def lrs_per_celltype(
    sample,
    sender = None,
    receiver = None,
    assay="raw",
    key="cci_scores",
    p_vals=None,
    n=15,
    x_label_size=18,
    y_label_size=24,
    x_tick_size=14,
    y_tick_size=12,
    figsize=(6, 5),
    show=True,
    title=None,
    title_size=14
):
    """Plots a bar plot of LR pairs and their proportions for a sender and receiver cell type pair along with p_values (optional).

    Args:
        sample (CCIData): The CCIData object.
        sender (str): The sender cell type. Defaults to None.
        receiver (str): The receiver cell type. Defaults to None.
        assay (str): The assay to use. Defaults to "raw".
        key (str): The key to use. Defaults to "cci_scores".
        p_vals (dict): A dictionary of p-values. Defaults to None.
        n (int): Number of LR pairs to plot. Defaults to 15.
        x_label_size (int): Font size for x-axis label. Defaults to 18.
        y_label_size (int): Font size for y-axis label. Defaults to 24.
        x_tick_size (int): Font size for tick labels. Defaults to 14.
        y_tick_size (int): Font size for tick labels. Defaults to 12.
        figsize (tuple): Size of the figure. Defaults to (10, 8).
        title (str) (optional): Title for the plot. Defaults to None.
        title_size (int): Font size of the title. Defaults to 14.

    Returns:
        matplotlib.figure.Figure: The figure
    """

    pairs = sample.get_lr_proportions(sender, receiver, assay, key)
    keys = list(pairs.keys())[:n]
    values = list(pairs.values())[:n]
    keys.reverse()
    values.reverse()
    if p_vals is not None:
        p_val_pairs = an.get_p_vals_per_celltype(p_vals, sender, receiver)
        labels = [p_val_pairs[key] for key in keys]

        # make labels readable (if less than 0.00001, show as <0.00001)
        for i in range(len(labels)):
            if labels[i] < 0.001:
                labels[i] = "<0.001"
            else:
                labels[i] = f"{labels[i]:.3f}"

        # Define colors based on p-values
        colors = [
            '#1f77b4' if val < 0.05 else 'grey' for val in [
                p_val_pairs[key] for key in keys]]

    # Create the figure and axis
    fig, ax = plt.subplots(figsize=figsize)
    plt.style.use('default')

    if p_vals is None:
        ax.barh(keys, values)
    else:
        bars = ax.barh(keys, values, color=colors)
        ax.bar_label(bars, labels)

    ax.set_xlabel("Proportion", fontsize=x_label_size)
    ax.set_ylabel("LR Pair", fontsize=y_label_size)
    ax.tick_params(axis='x', which='major', labelsize=x_tick_size)
    ax.tick_params(axis='y', which='major', labelsize=y_tick_size)

    if title:
        plt.title(title, pad=20, fontsize=title_size)
        
    plt.tight_layout()

    if show:
        plt.show()
    else:
        plt.close(fig)
        return fig



import pandas as pd
import numpy as np
from typing import Dict, List, Optional, Union

def align_dataframes(m1, m2, fill_value=0):
    """Aligns two DataFrames by matching their indices and columns, filling missing values with 0.

    Args:
        m1, m2 (pd.DataFrame): The DataFrames to align.

    Returns:
        tuple: A tuple of the aligned DataFrames.
    """

    m1, m2 = m1.align(m2, fill_value=fill_value)

    columns = sorted(set(m1.columns) | set(m2.columns))
    m1 = m1.reindex(columns, fill_value=fill_value)
    m2 = m2.reindex(columns, fill_value=fill_value)

    rows = sorted(set(m1.index) | set(m2.index))
    m1 = m1.reindex(rows, fill_value=fill_value)
    m2 = m2.reindex(rows, fill_value=fill_value)

    return m1, m2



import pandas as pd
import numpy as np
import networkx as nx
import scipy as sp
import scanpy
import gseapy as gp
import seaborn as sns
import statistics
import subprocess
import umap
import importlib.resources
import anndata as ad
import os

from matplotlib import pyplot as plt
from matplotlib.colors import LinearSegmentedColormap

from gseapy import barplot, dotplot
from tqdm import tqdm
from scipy.spatial.distance import pdist, squareform
from scipy.cluster.hierarchy import linkage, dendrogram, ClusterWarning
from scipy.cluster import hierarchy
from sklearn.cluster import AgglomerativeClustering, KMeans
from sklearn.metrics import silhouette_score, davies_bouldin_score
from sklearn import preprocessing as pp
from sklearn.decomposition import PCA
from warnings import simplefilter

from . import sc, pl, tl
from .CCIData_class import CCIData

simplefilter("ignore", Warning)


def calculate_dissim(
    sample1: CCIData,
    sample2: CCIData,
    assay: str = "raw",
    key: str = "cci_scores",
    lmbda: float = 0.5,
    ):
    """Calculates a dissimilarity score between two samples for each common LR pair.
    
    Args:
        sample1, sample2 (CCIData): The two samples to compare.
        assay (str) (optional): The assay to use for the comparison. Defaults to 'raw'.
        key (str) (optional): The key to use for the comparison. Defaults to  'cci_scores'.
        lmbda (float) (optional): The weighting factor for the comparison. Defaults to 0.5.

    Returns:
        dict: A dictionary where keys are common LR pairs and values are the dissimilarity scores.
    """
    
    if assay not in sample1.assays:
        raise ValueError(f"Assay {assay} not found in sample1.")
    if assay not in sample2.assays:
        raise ValueError(f"Assay {assay} not found in sample2.")
    
    if key not in sample1.assays[assay]:
        raise ValueError(f"Key {key} not found in sample1.")
    if key not in sample2.assays[assay]:
        raise ValueError(f"Key {key} not found in sample2.")
    
    dissims = {}
    
    for lr_pair in set(sample1.assays[assay][key].keys()).intersection(
        set(sample2.assays[assay][key].keys())
    ):
        dissims[lr_pair] = sc.dissimilarity_score(
            sample1.assays[assay][key][lr_pair],
            sample2.assays[assay][key][lr_pair],
            lmbda=lmbda
        )
        
    return dissims


def get_network_diff(
    network1: pd.DataFrame,
    network2: pd.DataFrame,
    perm_test: bool = True,
    num_perms: int = 100000
    ):
    """Calculates the difference between two networks. If perm_test is True, also performs permutation testing to assess the significance of the differences.

    Args:
        network1, network2 (pd.DataFrame): Two matrices to compare (as DataFrames).

    Returns:
        dict: A dictionary containing the difference between the two networks and p-values if perm_test is True.
    """

    dfs = tl.align_dataframes(network1, network2)
    diff = dfs[0] - dfs[1]
    p_vals = None
    
    if perm_test:
        abs_diff = abs(diff)

        result_matrix = abs_diff.values.copy()

        def permtr(x):
            return np.apply_along_axis(
                np.random.permutation,
                axis=0,
                arr=np.apply_along_axis(np.random.permutation, axis=1, arr=x),
            )

        # Permute and test for matrix1
        perm = [permtr(abs_diff) for _ in range(num_perms)]
        sums = np.sum([result_matrix < perm_result for perm_result in perm], axis=0)

        # Calculate p-values
        p_vals = sums / num_perms

        p_vals = pd.DataFrame(p_vals, index=abs_diff.index, columns=abs_diff.columns)
    
    return {"diff": diff, "p_vals": p_vals}
        

def cell_network_clustering(
    sample: CCIData, 
    assay: str = "raw",
    n_clusters: int = 0, 
    method: str = "KMeans"
    ):
    """Groups ligand-receptor (LR) pairs into clusters based on their interaction network similarity.
    
    This function:
    1. Separates LR pairs into those with single vs multiple interactions
    2. Calculates dissimilarity scores between all pairs of LR networks
    3. Performs clustering using either KMeans or Hierarchical clustering
    4. Determines optimal cluster number if not specified
    5. Creates new assays in the sample for each cluster

    Args:
        sample (CCIData): Sample containing LR interaction networks to cluster
        assay (str): Name of assay containing interaction data to analyze
        n_clusters (int): Number of desired clusters. If 0, determined automatically
        method (str): Clustering method - either "KMeans" or "Hierarchical"

    Returns:
        CCIData: Sample with new assays added for each cluster found

    Raises:
        ValueError: If specified assay not found in sample
    """

    if assay not in sample.assays:
        raise ValueError(f"Assay {assay} not found in sample.")

    # Dictionary to store LR pairs with only one interaction
    single_interaction_pairs = {}

    def has_non_zero_values(df):
        """Check if dataframe contains any non-zero values"""
        return not (df == 0).all().all()

    # Get all LR pairs with non-zero interactions
    multi_interaction_pairs = {
        key: value for key, value in sample.assays[assay]['cci_scores'].items()
        if has_non_zero_values(value)
    }
    
    # Separate out LR pairs with only a single interaction
    for lr_pair, interaction_matrix in list(multi_interaction_pairs.items()):
        # Check if matrix has only one non-zero value (excluding diagonal)
        diagonal_zeros = (interaction_matrix.values.diagonal() == 0).sum() 
        total_zeros = (interaction_matrix.values == 0).sum()
        matrix_size = interaction_matrix.shape[0]
        
        if (diagonal_zeros == matrix_size - 1 and 
            total_zeros == (matrix_size * matrix_size) - 1):
            single_interaction_pairs[lr_pair] = interaction_matrix
            del multi_interaction_pairs[lr_pair]

    # Process LR pairs with multiple interactions
    if multi_interaction_pairs:
        multi_dissimilarity_matrix = pd.DataFrame(
            index=multi_interaction_pairs.keys(), 
            columns=multi_interaction_pairs.keys()
        )
        
        print("Computing dissimilarity scores for complex interaction networks...")
        with tqdm(total=len(multi_interaction_pairs), desc="Processing") as pbar:
            for pair1, matrix1 in multi_interaction_pairs.items():
                for pair2, matrix2 in multi_interaction_pairs.items():
                    dissim_score = sc.dissimilarity_score(
                        matrix1, matrix2, lmbda=0.5, only_non_zero=True
                    )
                    multi_dissimilarity_matrix.loc[pair1, pair2] = dissim_score
                pbar.update(1)
                
        multi_interaction_clusters = _lr_cluster_helper(
            multi_dissimilarity_matrix, multi_interaction_pairs, n_clusters, method
        )

    # Process LR pairs with single interactions 
    if single_interaction_pairs:
        # Initialize an empty dataframe to store the results
        result_df_one = pd.DataFrame(
            index=single_interaction_pairs.keys(), columns=single_interaction_pairs.keys()
        )
        # Iterate through the keys and compare the dataframes
        print("Computing Dissimilarity Scores for single interactions...")
        with tqdm(total=len(single_interaction_pairs), desc="Processing") as pbar:
            for key1, df1 in single_interaction_pairs.items():
                for key2, df2 in single_interaction_pairs.items():
                    result = sc.dissimilarity_score(
                        df1, df2, lmbda=0.5, only_non_zero=True
                    )
                    
                    # Store the result in the result_df
                    result_df_one.loc[key1, key2] = result
                pbar.update(1)
                
        total = None
        for lr_pair in single_interaction_pairs.keys():
            if single_interaction_pairs[lr_pair].sum().sum() > 0:
                if total is not None:
                    total = total + single_interaction_pairs[lr_pair] / \
                    single_interaction_pairs[lr_pair].sum().sum()
                else:
                    total = single_interaction_pairs[lr_pair] / \
                        single_interaction_pairs[lr_pair].sum().sum()
                total = total.fillna(0)

        total = total / total.sum().sum()
        total = total.fillna(0)
        
        n_clusters = (
            total.values.diagonal()
            == 0
        ).sum()
        single_interaction_clusters = _lr_cluster_helper(
            result_df_one, single_interaction_pairs, n_clusters
        )
        single_interaction_clusters["cluster"] = (
            single_interaction_clusters["cluster"]
            + max(multi_interaction_clusters["cluster"])
            + 1
        )

    # Combine cluster assignments
    final_cluster_assignments = pd.concat(
        [multi_interaction_clusters, single_interaction_clusters]
    )

    # Create new assays for each cluster in the sample
    for lr_pair in final_cluster_assignments.index:
        cluster_id = final_cluster_assignments["cluster"][lr_pair]
        cluster_assay = f"cluster_{cluster_id}"
        
        # Initialize cluster assay if needed
        if cluster_assay not in sample.assays:
            sample.assays[cluster_assay] = {
                'cci_scores': {},
                'p_values': {}
            }
            
        # Add interaction matrices to cluster assay    
        sample.assays[cluster_assay]['cci_scores'][lr_pair] = \
            sample.assays[assay]['cci_scores'][lr_pair]
        sample.assays[cluster_assay]['p_values'][lr_pair] = \
            sample.assays[assay]['p_values'][lr_pair]

    # Calculate overall interactions for each cluster
    for assay_name in sample.assays:
        if assay_name.startswith('cluster_'):
            sample = sample.calc_overall(assay=assay_name)
    
    return sample


def _lr_cluster_helper(result_df, sample, n_clusters=0, method="KMeans"):
    """
    Args:
        result_df (pd.DataFrame): A DataFrame containing dissimilarity scores for LRs
        sample (dict): A dictionary containing LR matrices.
        n_clusters (int) (optional): The desired number of clusters. If 0, the optimal number is determined using silhouette analysis. Defaults to 0.
        method (str) (optional): The clustering method to use. Defaults to 'KMeans'.

    Returns:
        pd.DataFrame: A DataFrame with the cluster assignments for each sample.
    """

    # Compute distance matrix from disimilarity matrix
    result_df = result_df.astype("float64")
    result_df = result_df.fillna(0)
    distances = pdist(result_df.values, metric="euclidean")
    dist_matrix = squareform(distances)
    dist_matrix = pd.DataFrame(
        pp.MinMaxScaler(feature_range=(0, 100)).fit_transform(
            pd.DataFrame(dist_matrix).T.values
        )
    )

    print("Computing Principal Components of weighted graph ...")
    # Perform PCA on weighted edges of interaction network
    flatten_dfs = [df.to_numpy().flatten() for df in sample.values()]
    flatten_dfs = pd.DataFrame(flatten_dfs)
    flatten_dfs = flatten_dfs.fillna(0)
    scaler = pp.StandardScaler()
    data_scaled = scaler.fit_transform(flatten_dfs)
    pca = PCA(n_components=2)
    data_pca = pca.fit_transform(data_scaled)

    # Concatenate PC-1 and PC-2 with distance matrix
    pc_com_dist_matrix = pd.concat([dist_matrix, pd.DataFrame(data_pca)], axis=1)

    print("Performing Clustering and Ranking within clusters...")
    if n_clusters > 0:
        # Number of clusters (adjust as needed)
        n_clusters = n_clusters

        if method == "Hierarchial":
            # Perform hierarchical clustering
            model = AgglomerativeClustering(n_clusters=n_clusters, linkage="ward")
            clusters = model.fit_predict(pc_com_dist_matrix)
        if method == "KMeans":
            kmeans = KMeans(n_clusters=n_clusters, random_state=42)
            clusters = kmeans.fit_predict(pc_com_dist_matrix)

    if n_clusters == 0:
        if method == "Hierarchial":
            # Evaluate silhouette score for different numbers of clusters
            silhouette_scores = []
            for n_clusters in range(2, 11):
                clusterer = AgglomerativeClustering(
                    n_clusters=n_clusters, linkage="ward"
                )
                cluster_labels = clusterer.fit_predict(pc_com_dist_matrix)
                silhouette_avg = silhouette_score(pc_com_dist_matrix, cluster_labels)
                silhouette_scores.append(silhouette_avg)
            # Perform hierarchical clustering
            model = AgglomerativeClustering(
                # Add 2 to account for starting with k=2
                n_clusters=np.argmax(silhouette_scores) + 2,
                linkage="ward",
            )  # as indexing starts from 0
            clusters = model.fit_predict(pc_com_dist_matrix)

        if method == "KMeans":
            # Find optimal numer of clusters Davies-Bouldin index
            db_scores = []
            for k in range(2, 11):
                kmeans = KMeans(n_clusters=k, random_state=42)
                labels = kmeans.fit_predict(pc_com_dist_matrix)
                db_scores.append(davies_bouldin_score(pc_com_dist_matrix, labels))
            # Add 2 to account for starting with k=2
            optimal_clusters = np.argmin(db_scores) + 2
            kmeans = KMeans(n_clusters=optimal_clusters, random_state=42)
            clusters = kmeans.fit_predict(pc_com_dist_matrix)

    clusters = pd.DataFrame(clusters)
    clusters.index = sample.keys()
    clusters.rename(columns={0: "cluster"}, inplace=True)

    # Rank LRs in each cluster based on increasing dissimilarity
    dist_matrix.columns = list(sample.keys())
    dist_matrix.index = list(sample.keys())
    columns = ["LRs", "cluster"]
    final_clusters = pd.DataFrame(columns=columns)

    for i in range(0, len(set(clusters["cluster"]))):
        clusters_index = list(clusters[clusters["cluster"] == i].index)
        dist_matrix_cluster = dist_matrix[dist_matrix.index.isin(clusters_index)]
        similarity_matrix = 1 / (1 + dist_matrix_cluster)
        linkage_matrix = hierarchy.linkage(similarity_matrix, method="ward")
        sorted_indices = hierarchy.leaves_list(linkage_matrix)
        cluster_df = pd.DataFrame(clusters_index)
        cluster_df = cluster_df.iloc[sorted_indices[::-1]].reset_index(drop=True)
        cluster_df["cluster"] = i
        final_clusters = pd.concat([final_clusters, cluster_df])
        
    final_clusters = final_clusters.iloc[:, 1:]
    final_clusters = final_clusters.rename(columns={0: "LRs"})
    final_clusters = final_clusters[["LRs", "cluster"]]
    final_clusters.set_index("LRs", inplace=True)
    
    return final_clusters

    
def lr_interaction_clustering(
    sample,
    resolution=0.5,
    cluster_palette="Dark2_r",
    cell_type_palette="tab20",
    cell_colors=None,
    spot_size=1.5,
    spatial_plot=True, 
    proportion_plot=True,
    return_adata=False,
    **kwargs
    ):
    """Clustering of spatial LR interaction scores on AnnData objects processed through stLearn.

    Args:
        sample (AnnData): An AnnData object that has been run through stLearn.
        resolution (float) (optional): The resolution to use for the clustering. Defaults to 0.5.
        cluster_palette (str) (optional): Name of matplotlib colormap to use for clusters. Defaults to 'Dark2_r'.
        cell_type_palette (str) (optional): Name of matplotlib colormap to use for cell types. Defaults to 'tab20'.
        cell_colors (dict) (optional): Dictionary mapping cell types to colors. If not provided, colors will be generated from cell_type_palette. Defaults to None.
        spot_size (float) (optional): The size of the spots in the spatial plot. Defaults to 1.5.
        spatial_plot (bool) (optional): Whether to show the spatial plot. Defaults to True.
        proportion_plot (bool) (optional): Whether to show the proportion plot. Defaults to True.
        return_adata (bool) (optional): Whether to return the AnnData object with the clustering results. Defaults to False.
        **kwargs: Additional keyword arguments to pass to the scanpy spatial plot function.
    
    Returns:
        AnnData: An AnnData object with the clustering results.
    """

    if sample.adata is None:
        raise ValueError("No AnnData object found in sample.")
    
    LR = pd.DataFrame(sample.adata.obsm["lr_scores"])
    LR.columns = list(sample.adata.uns["lr_summary"].index)
    LR.index = sample.adata.obs.index

    LR = scanpy.AnnData(LR)
    scanpy.pp.normalize_total(LR, inplace=True)
    scanpy.pp.log1p(LR)
    scanpy.pp.pca(LR)
    scanpy.pp.highly_variable_genes(LR, flavor="seurat", n_top_genes=2000)
    scanpy.pp.neighbors(LR, use_rep="X_pca", n_neighbors=15)
    scanpy.tl.leiden(LR, resolution=resolution)
    
    scanpy.tl.rank_genes_groups(LR, groupby='leiden', method='wilcoxon')
    # scanpy.pl.rank_genes_groups(LR, n_genes=25, sharey=False)
    scanpy.tl.dendrogram(LR, groupby='leiden')
    scanpy.pl.rank_genes_groups_dotplot(LR, n_genes=10, groupby='leiden')
    
    LR.obsm = sample.adata.obsm
    LR.uns = sample.adata.uns
    LR.obs["leiden"] = LR.obs["leiden"].astype("int64")

    scanpy.pp.pca(LR)
    scanpy.pp.highly_variable_genes(LR, flavor="seurat", n_top_genes=2000)
    scanpy.pp.neighbors(LR, use_rep="X_pca", n_neighbors=15)
    scanpy.tl.umap(LR)
    
    sample.adata.obs["LR_Cluster"] = LR.obs["leiden"].astype("str")
    
    if spatial_plot:
        scanpy.pl.spatial(sample.adata, color="LR_Cluster", size=spot_size, 
                     palette=cluster_palette, **kwargs)
        
    if proportion_plot:
        if "cell_type" in sample.adata.uns:
            # make any value less than 0.1 == 0 in adata.uns['cell_type']
            sample.adata.uns['cell_type'] = sample.adata.uns['cell_type'].applymap(
                lambda x: 0 if x < 0.1 else x)
            
            merged_df = sample.adata.uns['cell_type'].merge(
                sample.adata.obs["LR_Cluster"], left_index=True, right_index=True)
            proportions = merged_df.groupby('LR_Cluster').mean()
            proportions = proportions.div(proportions.sum(axis=1), axis=0)
        else:
            proportions = sample.adata.obs.groupby('LR_Cluster')['cell_type'] \
                .value_counts(normalize=True).unstack()

        # Generate colors from cell_type_palette if cell_colors not provided 
        if cell_colors is None:
            unique_types = proportions.columns
            default_colors = plt.get_cmap(cell_type_palette).colors
            cell_colors = {ctype: default_colors[i % len(default_colors)] 
                        for i, ctype in enumerate(unique_types)}

        fig, (ax1, ax2) = plt.subplots(1, 2, gridspec_kw={'width_ratios': [4, 1]})

        bars = proportions.plot(kind="bar", stacked=True, ax=ax1, 
                                color=proportions.columns.to_series().map(cell_colors))  
        # Use colormap based on cell types
        ax1.set_ylabel("Proportion")
        ax1.legend().remove()  # Remove default legend

        # Create a legend without plot elements
        handles, labels = bars.get_legend_handles_labels()  # Get handles and labels 
        # from the bar plot
        ax2.legend(handles, labels, loc='center')
        ax2.axis('off')

        plt.tight_layout()
        plt.show()

    if return_adata:
        return LR


def run_gsea(
    sample: CCIData = None,
    assay: str = "raw",
    lrs=None,
    organism="human",
    gene_sets=["KEGG_2021_Human", "MSigDB_Hallmark_2020"],
    show_dotplot=False,
    show_barplot=True,
    top_term=5,
    figsize=(3,5),
    use_background=True,
    background_list=None,
    return_results=True
):
    """Runs GSEA analysis on a sample.

    Args:
        sample (CCIData) (optional): The sample to run GSEA on. If not given, lrs are used. Defaults to None.
        assay (str) (optional): The assay to use for the GSEA analysis. Defaults to 'raw'.
        lrs (list) (optional): A list of LR pairs to use for GSEA analysis instead of sample. Defaults to None.
        organism (str) (optional): The organism to use. Defaults to 'human'.
        gene_sets (list) (optional): The gene sets to use for gseapy analysis. Defaults to ['KEGG_2021_Human', 'MSigDB_Hallmark_2020'].
        show_dotplot (bool) (optional): Whether to show the dotplot. Defaults to False.
        show_barplot (bool) (optional): Whether to show the barplot. Defaults to True.
        top_term (int) (optional): The number of top terms to show. Defaults to 5.
        figsize (tuple) (optional): The size of the figure. Defaults to (3,5).
        use_background (bool) (optional): Whether to use a background list. Defaults to True.
        background_list (list) (optional): A list of genes to use as background. If none given, then ConnectomeDB2020_lit genes are used. Defaults to None.
        return_results (bool) (optional): Whether to return the results DataFrame.  Defaults to True.

    Returns:
        pd.DataFrame or None: A DataFrame with the GSEA results if return_results=True, otherwise None.
    """

    gene_list = set()

    if lrs is None:
        if assay not in sample.assays:
            raise ValueError(f"Assay {assay} not found in sample.")
        lrs = sample.assays[assay]["cci_scores"].keys()

    for lr in lrs:
        gene1, gene2 = lr.split("_")
        gene_list.add(gene1)
        gene_list.add(gene2)

    gene_list = list(gene_list)
    
    background = None
    
    if use_background:
        if background_list is not None:
            background = background_list
        else:
            path = os.path.dirname(os.path.realpath(__file__))
            background = pd.read_csv(f"{path}/data/connectomeDB2020_lit.txt", sep="\t", header=None)
            background = list(set(background[0].tolist() + background[1].tolist()))

    enr = gp.enrichr(
        gene_list=gene_list,
        gene_sets=gene_sets,
        organism=organism,
        outdir=None,
        background=background
    )
        
    if show_dotplot:
        try:
            ax = dotplot(
                enr.results,
                column="Adjusted P-value",
                x="Gene_set",
                size=10,
                top_term=top_term,
                figsize=figsize,
                xticklabels_rot=45,
                show_ring=True,
                marker="o",
            )
        except:
            print("Could not plot dotplot.")
        
    if show_barplot:
        colours = list(plt.cm.Dark2.colors)
        colour_dict = {gene_set: colours[i] for i, gene_set in enumerate(gene_sets)}
    
        try:
            ax = barplot(
                enr.results,
                column="Adjusted P-value",
                group="Gene_set",
                size=10,
                top_term=top_term,
                figsize=figsize,
                color=colour_dict
            )
        except:
            print("Could not plot barplot.")

    if return_results:
        return enr.results
    return None


def pathway_subset(
    sample: CCIData,
    assay: str = "raw", 
    gsea_results: pd.DataFrame = None, 
    terms: list = None, 
    strict: bool = False,
    assay_name: str = None
    ) -> CCIData:
    """Subsets a sample to only include interactions between genes in a set of pathways.

    Args:
        sample (CCIData): The sample to subset.
        assay (str) (optional): The assay to use for the subset. Defaults to 'raw'.
        gsea_results (pd.DataFrame): The GSEA results to use to subset the sample.
        terms (list): The terms to subset the sample with.
        strict (bool): Whether to only include interactions between genes in the same pathway.

    Returns:
        CCIdata: The sample with an added assay containing the subsetted interactions.
    """

    genes = []
    grouped_cci_scores = {}
    grouped_p_values = {}
    
    if assay_name is None:
        # make assay name based on terms
        assay_name = "+".join(terms)    
        
    if assay not in sample.assays:
        raise ValueError(f"Assay {assay} not found in sample.")

    for term in terms:
        if term not in gsea_results['Term'].tolist():
            raise ValueError(f"Term {term} not found in GSEA results.")
        filtered_df = gsea_results[gsea_results['Term'] == term]
        gene_list = filtered_df['Genes'].tolist()

        for gene in gene_list:
            genes.extend(gene.lower().split(";"))

    for key in sample.assays[assay]['cci_scores'].keys():
        lig, rec = key.lower().split("_")
        if strict:
            if lig in genes and rec in genes:
                grouped_cci_scores[key] = sample.assays[assay]['cci_scores'][key]
                grouped_p_values[key] = sample.assays[assay]['p_values'][key]
        else:
            if lig in genes or rec in genes:
                grouped_cci_scores[key] = sample.assays[assay]['cci_scores'][key]
                grouped_p_values[key] = sample.assays[assay]['p_values'][key]
                
    print(f"Number of interactions in {assay_name}: {len(grouped_cci_scores)}")

    sample.assays[assay_name] = {}
    sample.assays[assay_name]['cci_scores'] = grouped_cci_scores
    sample.assays[assay_name]['p_values'] = grouped_p_values
    sample = sample.calc_overall(assay_name)
    
    return sample


def add_lr_module_score(sample, lr_list, key_name="score"):
    """Adds a module score to an AnnData object run through stLearn based on the interactions in a list of ligand-receptor pairs.

    Args:
        sample (AnnData): The AnnData object to add the score to. Must be processed through stLearn.
        lr_list (list): The list of ligand-receptor pairs to use.
        key_name (str): The key to use for the score.

    Returns:
        AnnData: The AnnData object with the module score added.
    """

    lr_counts = pd.DataFrame(sample.obsm['lr_sig_scores'])
    lr_counts.index = sample.obs.index
    lr_counts.columns = sample.uns['lr_summary'].index

    adata = ad.AnnData(lr_counts)
    scanpy.tl.score_genes(adata, gene_list=lr_list)
    sample.obs[key_name] = adata.obs['score']

    return sample



import pandas as pd
import numpy as np
import scanpy
import pickle
import json
from .CCIData_class import CCIData

def read_stLearn(path, key="cell_type", save_anndata=True) -> CCIData:
    """Reads a stLearn ligand-receptor analysis output and converts it to a CCIData object.

    Args:
        path (str): The path to the stLearn ligand-receptor analysis output.
        key (str) (optional): The key in adata.obs that was used for CCI. Defaults to "cell_type".
        save_anndata (bool) (optional): Whether to save the AnnData object in the CCIData object. Defaults to True.

    Returns:
        CCIData: The CCIData object.
    """

    adata = scanpy.read_h5ad(path)

    if save_anndata:
        cci_data = CCIData(
            n_spots=adata.shape[0], 
            cci_scores=adata.uns[f"per_lr_cci_raw_{key}"], 
            p_values=adata.uns[f"per_lr_cci_pvals_{key}"],
            adata=adata
            )
    else:
        cci_data = CCIData(
            n_spots=adata.shape[0], 
            cci_scores=adata.uns[f"per_lr_cci_raw_{key}"], 
            p_values=adata.uns[f"per_lr_cci_pvals_{key}"]
            )

    return cci_data


def convert_stLearn(adata, key="cell_type", save_anndata=True) -> CCIData:
    """Reads a stLearn ligand-receptor analysis output and converts it to a CCIData object.

    Args:
        adata (AnnData): The stLearn ligand-receptor analysis output.
        key (str) (optional): The key in adata.obs that was used for CCI. Defaults to "cell_type".
        save_anndata (bool) (optional): Whether to save the AnnData object in the CCIData object. Defaults to True.

    Returns:
        CCIData: The CCIData object.
    """

    if save_anndata:
        cci_data = CCIData(
            n_spots=adata.shape[0], 
            cci_scores=adata.uns[f"per_lr_cci_raw_{key}"], 
            p_values=adata.uns[f"per_lr_cci_pvals_{key}"],
            adata=adata
            )
    else:
        cci_data = CCIData(
            n_spots=adata.shape[0], 
            cci_scores=adata.uns[f"per_lr_cci_raw_{key}"], 
            p_values=adata.uns[f"per_lr_cci_pvals_{key}"]
            )

    return cci_data


def read_CellPhoneDB(means_path, pvals_path, n_spots=None) -> CCIData:
    """Reads a CellPhoneDB interaction scores txt file and converts it to a CCIDataobject.

    Args:
        means_path (str): Path to the means txt file.
        pvals_path (str): Path to the pvals txt file.
        n_spots (int) (optional): The number of spots. Defaults to None.

    Returns:
        CCIData: The CCIData object.
    """

    cp_obj = pd.read_csv(means_path, delimiter="\t")
    lr_dict = {}
    for ind in cp_obj.index:

        key = cp_obj["interacting_pair"][ind]

        val = cp_obj.iloc[ind, cp_obj.columns.get_loc("classification") + 1:]
        val = pd.DataFrame({"Index": val.index, "Value": val.values})
        val[["Row", "Col"]] = val["Index"].str.split("|", expand=True)
        val = val.drop("Index", axis=1)
        val = val.pivot(index="Row", columns="Col", values="Value")
        val = val.rename_axis(None, axis=1).rename_axis(None)
        lr_dict[key] = val

    cp_obj = pd.read_csv(pvals_path, delimiter="\t")
    p_vals_dict = {}
    for ind in cp_obj.index:

        key = cp_obj["interacting_pair"][ind]

        val = cp_obj.iloc[ind, cp_obj.columns.get_loc("classification") + 1:]
        val = pd.DataFrame({"Index": val.index, "Value": val.values})
        val[["Row", "Col"]] = val["Index"].str.split("|", expand=True)
        val = val.drop("Index", axis=1)
        val = val.pivot(index="Row", columns="Col", values="Value")
        val = val.rename_axis(None, axis=1).rename_axis(None)
        val = val.fillna(1)
        p_vals_dict[key] = val

    cci_data = CCIData(n_spots=n_spots, cci_scores=lr_dict, p_values=p_vals_dict)

    return cci_data


def read_Squidpy(result, n_spots=None) -> CCIData:
    """Reads a Squidpy ligand-receptor analysis output and converts it to a CCIData
    object.

    Args:
        result (dict): The output from squidpy.gr.ligrec.
        n_spots (int) (optional): The number of spots. Defaults to None.

    Returns:
        CCIData: The CCIData object.
    """

    lr_dict = {}
    pvals = pd.DataFrame(result["pvalues"])

    cci_names = np.array([col[0] + "--" + col[1] for col in pvals.columns])
    cell_type_set = np.unique([col[0] for col in pvals.columns])

    for i, row in enumerate(pvals.index):

        int_matrix = np.zeros((len(cell_type_set), len(cell_type_set)))
        lr_ = "_".join(list(row))
        lr_ccis = cci_names
        lig = list(row)[0]
        rec = list(row)[1]
        for j, cci in enumerate(lr_ccis):
            c1, c2 = cci.split("--")
            row = np.where(cell_type_set == c1)[0][0]
            col = np.where(cell_type_set == c2)[0][0]
            int_matrix[row, col] = result["means"][c1][c2][lig][rec]

        lr_dict[lr_] = pd.DataFrame(
            int_matrix, index=cell_type_set, columns=cell_type_set
        )

    p_vals_dict = {}

    for i, row in enumerate(pvals.index):

        int_matrix = np.zeros((len(cell_type_set), len(cell_type_set)))
        lr_ = '_'.join(list(row))
        lr_ccis = cci_names
        lig = list(row)[0]
        rec = list(row)[1]
        for j, cci in enumerate(lr_ccis):
            c1, c2 = cci.split('--')
            row = np.where(cell_type_set == c1)[0][0]
            col = np.where(cell_type_set == c2)[0][0]
            int_matrix[row, col] = result['pvalues'][c1][c2][lig][rec]

        df = pd.DataFrame(int_matrix, index=cell_type_set, columns=cell_type_set)
        df = df.fillna(1)
        p_vals_dict[lr_] = df

    cci_data = CCIData(n_spots=n_spots, cci_scores=lr_dict, p_values=p_vals_dict)

    return cci_data


def read_CellChat(path, n_spots=None) -> CCIData:
    """Reads a CellChat ligand-receptor analysis output (cellchat@dr) and converts it to
    a CCIData object.

    Args:
        path (str): The output from cellchat@dr.
        n_spots (int) (optional): The number of spots. Defaults to None.

    Returns:
        CCIData: The CCIData object.
    """

    result = pd.read_csv(path)
    lr_dict = {}
    pvals_dict = {}

    cell_type_set = np.unique(np.concatenate([result["source"], result["target"]]))

    for i in result.index:
        lr_ = result['interaction_name'][i]
        if lr_ not in lr_dict.keys():
            int_matrix = np.zeros((len(cell_type_set), len(cell_type_set)))
        else:
            int_matrix = lr_dict[lr_].values
        row = np.where(cell_type_set == result["source"][i])[0][0]
        col = np.where(cell_type_set == result["target"][i])[0][0]
        int_matrix[row, col] = result["prob"][i]

        lr_dict[lr_] = pd.DataFrame(
            int_matrix, index=cell_type_set, columns=cell_type_set
        )

    for i in result.index:
        lr_ = result['interaction_name'][i]
        if lr_ not in pvals_dict.keys():
            int_matrix = np.ones((len(cell_type_set), len(cell_type_set)))
        else:
            int_matrix = pvals_dict[lr_].values
        row = np.where(cell_type_set == result["source"][i])[0][0]
        col = np.where(cell_type_set == result["target"][i])[0][0]
        int_matrix[row, col] = result["pval"][i]

        df = pd.DataFrame(
            int_matrix, index=cell_type_set, columns=cell_type_set
        )
        df = df.fillna(1)
        pvals_dict[lr_] = df

    cci_data = CCIData(n_spots=n_spots, cci_scores=lr_dict, p_values=pvals_dict)

    return cci_data


def read_NATMI(path, n_spots=None) -> CCIData:
    """Reads a NATMI ligand-receptor analysis output (Edges_lrc2p.csv) and converts it
    to a CCIData object.
    Args:
        path (str): The path to Edges_lrc2p.csv.
        n_spots (int) (optional): The number of spots. Defaults to None.

    Returns:
        CCIData: The CCIData object.
    """

    result = pd.read_csv(path)
    lr_dict = {}

    cell_type_set = np.unique(np.concatenate(
        [result["Sending cluster"], result["Target cluster"]]))

    for i in result.index:
        lr_ = result['Ligand symbol'][i] + "_" + result['Receptor symbol'][i]
        if lr_ not in lr_dict.keys():
            int_matrix = np.zeros((len(cell_type_set), len(cell_type_set)))
        else:
            int_matrix = lr_dict[lr_].values
        row = np.where(cell_type_set == result["Sending cluster"][i])[0][0]
        col = np.where(cell_type_set == result["Target cluster"][i])[0][0]
        int_matrix[row, col] = result["Edge average expression weight"][i]

        lr_dict[lr_] = pd.DataFrame(
            int_matrix, index=cell_type_set, columns=cell_type_set
        )

    cci_data = CCIData(n_spots=n_spots, cci_scores=lr_dict)
    
    return cci_data


def read_CCIData(path: str) -> CCIData:
    """Loads a CCIData object from JSON or pickle file.
    
    Args:
        path (str): Path to the saved CCIData file
        
    Returns:
        CCIData: The loaded CCIData object
    """
    if path.endswith('.json'):
        # Load from JSON file
        with open(path, 'r') as f:
            data_dict = json.load(f)
        return from_dict(data_dict)
    
    elif path.endswith('.pkl'):
        # Load from pickle file for backwards compatibility
        with open(path, 'rb') as f:
            return pickle.load(f)
    else:
        raise ValueError("File must be .json or .pkl format")
    
    
def read_network(path: str, metadata: dict = None) -> CCIData:
    """Loads a single network from a csv file.
    
    Args:
        path (str): Path to the saved network file
        metadata (dict) (optional): Dictionary containing other metadata
        
    Returns:
        CCIData: The loaded CCIData object
    """
    if path.endswith('.csv'):
        # Load from CSV file
        data = pd.read_csv(path, index_col=0)
        return CCIData(network=data, other_metadata=metadata)
    
    else:
        raise ValueError("File must be .csv format")


def from_dict(data_dict: dict) -> CCIData:
    """Create CCIData object from a dictionary.
    
    Args:
        data_dict: Dictionary containing CCIData representation
        
    Returns:
        CCIData: Reconstructed CCIData object
    """
    # Convert assays back to DataFrames
    assays = {}
    for assay_name, assay in data_dict['assays'].items():
        assays[assay_name] = {}
        for key, value in assay.items():
            if key in ['cci_scores', 'p_values']:
                # Convert dict of DataFrames
                assays[assay_name][key] = {
                    k: pd.DataFrame.from_dict(v) for k,v in value.items()
                }
            elif isinstance(value, dict):
                # Convert single DataFrame
                assays[assay_name][key] = pd.DataFrame.from_dict(value)
            else:
                assays[assay_name][key] = value

    return CCIData(
        other_metadata=data_dict['metadata'],
        assays=assays
    )
    


import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
import copy

from scipy import stats
from tqdm import tqdm

from . import sc, tl
from .CCIData_class import CCIData


def get_lr_pairs(
    samples: list,
    assay = "raw",
    method = ">=50%"
    ) -> list:
    """Identifies the LR pairs present in a list of samples according to the given method.

    Args:
        samples (list): A list of CCIData objects.
        assay (str) (optional): The assay or assays to use for identifying LR pairs. Defaults to "raw".
        method (str) (optional): The method to use for identifying LR pairs. Options are "intersection", ">=50%", ">50%", and "union". Defaults to ">=50%".

    Returns:
        list: A list of LR pairs that are present in a majority of samples
    """

    lr_pairs_counts = {}
    lr_pairs = []
    assays = []
    
    if type(assay) == str:
        assays = [assay for i in range(len(samples))]
    elif len(assay) != len(samples):
        raise ValueError("Assay list must be the same length as the samples list.")
    else:
        assays = assay

    for i in range(len(samples)):
        for lr_pair, matrix in samples[i].assays[assays[i]]['cci_scores'].items():
            if sum(sum(matrix.values)) != 0:
                lr_pairs_counts[lr_pair] = lr_pairs_counts.setdefault(lr_pair, 0) + 1

    for lr_pair, count in lr_pairs_counts.items():
        if method == "intersection":
            if count == len(samples):
                lr_pairs.append(lr_pair)
        elif method == ">=50%":
            if count >= len(samples) / 2:
                lr_pairs.append(lr_pair)
        elif method == ">50%":
            if count > len(samples) / 2:
                lr_pairs.append(lr_pair)
        elif method == "union":
            lr_pairs.append(lr_pair)
        else:
            raise ValueError("Method must be 'intersection', '>=50%', '>50%', or 'union'.")

    return lr_pairs


def calc_scale_factors(
    samples, 
    method="mean", 
    assay="raw", 
    group_key="platform"
    ) -> dict:
    """Calculates the scale factors for normalizing matrices between different platforms.
    
    Args:
        samples (list): A list of CCIData objects.
        method (str) (optional): The method to use for calculating scale factors.
        assay (str) (optional): The assay or assays to use for calculating scale factors. Defaults to "raw".
        group_key (str) (optional): The key to use for grouping samples. Defaults to "platform".
        
    Returns:
        dict: A dictionary where keys are LR pairs and values are the scale factors.
    """
    
    scale_factors = {}
    assays = []
    
    if type(assay) == str:
        assays = [assay for i in range(len(samples))]
    elif len(assay) != len(samples):
        raise ValueError("Assay list must be the same length as the samples list.")
    else:
        assays = assay
    
    # get all the unique group keys
    group_keys = set([sample.metadata[group_key] for sample in samples])
    group_samples = {group_key: [] for group_key in group_keys}
    
    for sample in samples:
        group_samples[sample.metadata[group_key]].append(sample)
    

    # get the scale factors for each group
    for group, sample_list in group_samples.items():
        total_counts = 0
        lr_pair_count = 0
        for i in range(len(sample_list)):
            for lr_pair, df in sample_list[i].assays[assays[i]]['cci_scores'].items():
                lr_pair_count += 1
                if method == "mean":
                    total_counts += df.mean().mean()
                elif method == "max":
                    total_counts += df.max().max()
                elif method == "median":
                    total_counts += df.median().median()
                elif method == "sum":
                    total_counts += df.sum().sum()
                else:
                    raise ValueError("Invalid method option.")

        total_counts = total_counts / lr_pair_count
        scale_factors[group] = total_counts
        
    # divide each group by the max scale factor
    max_scale_factor = max(scale_factors.values())
    for group, scale_factor in scale_factors.items():
        scale_factors[group] = max_scale_factor / scale_factor
        
    return scale_factors


def lr_integration(
    samples,
    method=">=50%",
    sum=False,
    strict=False,
    assay="raw",
    integrate_pvals=True,
    p_val_method="stouffer",
    metadata=None
    ) -> CCIData:
    """Integrates a list of samples into a single sample per lr pair.
    
    Args:
        samples (list): A list of CCIData objects.
        method (str) (optional): The method to use for identifying LR pairs. Options are "intersection", ">=50%", ">50%", and "union". Defaults to ">=50%".
        sum (bool) (optional): Whether to sum instead of multiply the matrices. Defaults to False.
        strict (bool) (optional): If True, only interactions where more than 50% of the values are non-zero will be multiplied. Defaults to False.
        assay (str) (optional): The assay or assays to use for integrating samples. Defaults to "raw".
        integrate_pvals (bool) (optional): Whether to integrate p-values (if possible). Defaults to True.
        p_val_method (str) (optional): The method to use for combining p-values. Options are "stouffer" and "fisher". Defaults to "stouffer".
        metadata (dict) (optional): Additional metadata to include in the integrated sample. Defaults to None.
        
    Returns:
        CCIData: The integrated sample.
    """
    
    if len(samples) == 0:
        raise ValueError("No samples provided.")
    
    assays = []
    
    if type(assay) == str:
        assays = [assay for i in range(len(samples))]
    elif len(assay) != len(samples):
        raise ValueError("Assay list must be the same length as the samples list.")
    else:
        assays = assay
    
    for i in range(len(samples)):
        if samples[i].assays[assays[i]] is None:
            raise ValueError("Sample does not have the specified assay.")
        
        if integrate_pvals:
            if 'p_values' not in samples[i].assays[assays[i]]:
                integrate_pvals = False
                print("No p-values found. Skipping p-value integration.")
                
    integrated_cci_scores = {}
    lr_dfs = {}

    if len(samples) >= 2:
        lr_pairs = sorted(get_lr_pairs(samples, assay=assays, method=method))
    else:
        raise ValueError("Integration needs at least two samples")

    for i in range(len(lr_pairs)):
        lr = lr_pairs[i]

        for j in range(len(samples)):
            if lr in samples[j].assays[assays[j]]['cci_scores']:
                if lr in lr_dfs:
                    lr_dfs[lr].append(samples[j].assays[assays[j]]['cci_scores'][lr])
                else:
                    lr_dfs[lr] = [samples[j].assays[assays[j]]['cci_scores'][lr]]

    with tqdm(total=len(lr_pairs), desc="Integrating LR CCI scores") as pbar:
        for lr, dfs in lr_dfs.items():
            if len(dfs) == 2:
                if sum:
                    dfs[0], dfs[1] = tl.align_dataframes(dfs[0], dfs[1])
                    integrated_cci_scores[lr] = dfs[0] + dfs[1]
                    integrated_cci_scores[lr] = integrated_cci_scores[lr] / 2
                    integrated_cci_scores[lr] = integrated_cci_scores[lr].fillna(0)
                else:
                    integrated_cci_scores[lr] = (dfs[0] * dfs[1]).fillna(0)
                    integrated_cci_scores[lr] = \
                        np.sqrt(integrated_cci_scores[lr]).fillna(0)
                        
            elif len(dfs) > 2:
                if sum:
                    integrated_cci_scores[lr] = dfs[0]
                    for i in range(1, len(dfs)):
                        integrated_cci_scores[lr], dfs[i] = tl.align_dataframes(
                            integrated_cci_scores[lr], dfs[i])
                        integrated_cci_scores[lr] = \
                            integrated_cci_scores[lr] + dfs[i]
                    integrated_cci_scores[lr] = \
                        integrated_cci_scores[lr] / len(dfs)
                    integrated_cci_scores[lr] = integrated_cci_scores[lr].fillna(0)
                else:
                    integrated_cci_scores[lr] = sc.multiply_non_zero_values(dfs, 
                                                                 strict=strict)
                    
            else:
                integrated_cci_scores[lr] = dfs[0]
            tqdm.update(pbar, 1)

    integrated_p_values = None
    
    if integrate_pvals:
        integrated_p_values = {}
        lr_dfs = {}

        for i in range(len(lr_pairs)):
            lr = lr_pairs[i]

            for j in range(len(samples)):
                if lr in samples[j].assays[assays[j]]['p_values']:
                    if lr in lr_dfs:
                        lr_dfs[lr].append(samples[j].assays[assays[j]]['p_values'][lr])
                    else:
                        lr_dfs[lr] = [samples[j].assays[assays[j]]['p_values'][lr]]

        with tqdm(total=len(lr_dfs), desc="Integrating p values") as pbar:
            for lr, dfs in lr_dfs.items():
                integrated_p_values[lr] = \
                    _correct_pvals_matrix(dfs, method=p_val_method)
                pbar.update(1)
                
    if metadata is None:
        metadata = {'platform': 'integrated'}
        
    integrated = CCIData(
        cci_scores=integrated_cci_scores,
        p_values=integrated_p_values,
        other_metadata=metadata
    )
    
    return integrated


def integrate_networks(
    samples,
    sum=False,
    strict=False,
    assay="raw",
    network_key="network",
    integrate_pvals=False,
    p_val_method="stouffer",
    p_val_key="network_p_values",
    ) -> CCIData:
    """Integrates a list of samples that are single networks into a single network.
    
    Args:
        samples (list): A list of CCIData objects.
        sum (bool) (optional): Whether to sum instead of multiply the matrices. Defaults to False.
        strict (bool) (optional): If True, only interactions where more than 50% of the  values are non-zero will be multiplied. Defaults to False.
        assay (str) (optional): The assay or assays to use for integrating samples.  Defaults to "raw".
        network_key (str) (optional): The key or keys to use for identifying the  network. Defaults to "network".
        integrate_pvals (bool) (optional): Whether to integrate p-values (if possible). Defaults to False.
        p_val_method (str) (optional): The method to use for combining p-values. Options are "stouffer" and "fisher". Defaults to "stouffer".
        p_val_key (str) (optional): The key to use for identifying p-values. Defaults to "network_p_values".
        
    Returns:
        CCIData: The integrated sample.
    """
    
    if len(samples) < 2:
        raise ValueError("Integration needs at least two samples.")
    
    networks = []
    integrated_network = None
    integrated_p_values = None
    assays = []
    network_keys = []
    
    if type(assay) == str:
        assays = [assay for i in range(len(samples))]
    elif len(assay) != len(samples):
        raise ValueError("Assay list must be the same length as the samples list.")
    else:
        assays = assay
        
    if type(network_key) == str:
        network_keys = [network_key for i in range(len(samples))]
    elif len(network_key) != len(samples):
        raise ValueError("Network key list must be the same length as the samples list.")
    else:
        network_keys = network_key
    
    for i in range(len(samples)):
        if samples[i].assays[assays[i]] is None:
            raise ValueError("Sample does not have the specified assay.")
        
        if integrate_pvals:
            if p_val_key is None:
                raise ValueError("No p-value key provided.")
            
            if p_val_key not in samples[i].assays[assays[i]]:
                raise ValueError("Sample does not have the specified p-value key.")
            
        networks.append(samples[i].assays[assays[i]][network_keys[i]])
            
    if len(networks) == 2:
        if sum:
            networks[0], networks[1] = tl.align_dataframes(networks[0], networks[1])
            integrated_network = networks[0] + networks[1]
            integrated_network = integrated_network / 2
            integrated_network = integrated_network.fillna(0)
        else:
            integrated_network = (networks[0] * networks[1]).fillna(0)
            integrated_network = np.sqrt(integrated_network).fillna(0)
            
    elif len(networks) > 2:
        if sum:
            integrated_network = networks[0]
            for i in range(1, len(networks)):
                integrated_network, networks[i] = tl.align_dataframes(
                    integrated_network, networks[i])
                integrated_network = integrated_network + networks[i]
            integrated_network = integrated_network / len(networks)
            integrated_network = integrated_network.fillna(0)
        else:
            integrated_network = sc.multiply_non_zero_values(networks, strict=strict)

    if integrate_pvals:
        p_values = []
        
        for i in range(len(samples)):
            p_values.append(samples[i].assays[assays[i]][p_val_key])
        
        integrated_p_values = _correct_pvals_matrix(p_values, method=p_val_method)
        
    integrated = CCIData(
        network=integrated_network,
        network_p_values=integrated_p_values
    )

    return integrated


def _correct_pvals_matrix(
    dfs, 
    method="stouffer"
    ):
    
    result_df = dfs[0]

    for i in range(len(dfs)):
        dfs[i], result_df = tl.align_dataframes(
            dfs[i], result_df, fill_value=np.NaN)

    for i in range(len(dfs)):
        dfs[i], result_df = tl.align_dataframes(
            dfs[i], result_df, fill_value=np.NaN)

    result_df = result_df.astype(np.float64)
    for i, row in result_df.iterrows():
        for j in row.index:
            values = [df.loc[i, j] for df in dfs]
            values = [
                0.00000000001 if x == 0 else (
                    0.9999999999 if x == 1 else x) for x in values]
            values = [x for x in values if not np.isnan(x)]
            result_df.loc[i, j] = stats.combine_pvalues(values, method=method)[1]

    return result_df



import numpy as np

from . import tl


def dissimilarity_score(
    m1,
    m2,
    lmbda=0.5,
    normalise=False,
    binary=False,
    trim=False,
    only_non_zero=False
):
    """Calculates a dissimilarity score between two matrices.

    Args:
        m1, m2 (pd.DataFrame): Two matrices to compare.
        lmbda (float) (optional): Weighting factor for weighted vs binary dissimilarity (0-1). 0 is fully binary and 1 is fully weighted. Defaults to 0.5.
        normalise (bool) (optional): Normalizes matrices before comparison. Defaults to False.
        binary (bool) (optional): Treats matrices as binary (0 or 1). Defaults to False.
        trim (bool) (optional): Trims matrices to common rows and columns. Otherwise pads 0s to uncommon rows and columns. Defaults to False.
        only_non_zero (bool) (optional): Only considers non-zero edges for calculation. Defaults to False.

    Returns:
        pd.DataFrame: The dissimilarity scores between the two matrices.
    """

    if trim:
        common_rows = list(set(m1.index) & set(m2.index))
        common_cols = list(set(m1.columns) & set(m2.columns))

        m1 = m1.loc[common_rows, common_cols]
        m2 = m2.loc[common_rows, common_cols]

    else:
        m1, m2 = tl.align_dataframes(m1, m2)

    m1 = m1.values
    m2 = m2.values

    if normalise:
        if m1.sum().sum() == 0 and m2.sum().sum() == 0:
            return 0
        if m1.sum().sum() != 0 and m2.sum().sum() != 0:
            m1 = m1 / m1.sum().sum()
            m2 = m2 / m2.sum().sum()

    if binary:
        m1 = np.where(m1 > 0, 1, 0)
        m2 = np.where(m2 > 0, 1, 0)

    n_of_edges = len(m1) ** 2

    if only_non_zero:
        n_of_edges = np.where((m1 + m2) > 0, 1, 0).sum().sum()

    abs_weight_difference = np.abs(m1 - m2)
    weight_sum = m1 + m2

    # Avoid division by zero
    weight_sum[weight_sum == 0] = -1
    norm_weight_difference = abs_weight_difference / weight_sum
    norm_weight_difference_sum = np.sum(np.sum(norm_weight_difference))
    wt_dissim = lmbda * (norm_weight_difference_sum / n_of_edges)

    n_diff = np.where(abs_weight_difference > 0, 1, 0).sum().sum()
    bin_dissim = (1 - lmbda) * (n_diff / n_of_edges)

    return wt_dissim + bin_dissim


def multiply_non_zero_values(dataframes, strict=False):
    """Multiply non-zero values across a list of pandas DataFrames.

    Args:
    dataframes (list): A list of pandas DataFrames with the same shape and column/row names.
    strict (bool) (optional): If True, only interactions where more than 50% of the values are non-zero will be multiplied. Defaults to False.

    Returns:
        pd.DataFrame: A new DataFrame where each cell contains the product of non-zero values or zero if more than 50% of the values in the corresponding cells are zero.
    """

    result_df = dataframes[0]

    for i in range(len(dataframes)):
        dataframes[i], result_df = tl.align_dataframes(dataframes[i], result_df)

    for i in range(len(dataframes)):
        dataframes[i], result_df = tl.align_dataframes(dataframes[i], result_df)

    result_df = result_df.astype(np.float64)
    for i, row in result_df.iterrows():
        for j in row.index:
            values = [df.loc[i, j] for df in dataframes]
            non_zero_values = [value for value in values if value != 0]

            if strict:
                if len(non_zero_values) / len(values) <= 0.5:
                    result_df.loc[i, j] = 0
                else:
                    result_df.loc[i, j] = np.prod(non_zero_values, dtype=np.float64)
            else:
                if len(non_zero_values) == 0:
                    result_df.loc[i, j] = 0
                else:
                    result_df.loc[i, j] = np.prod(non_zero_values, dtype=np.float64)

    result_df = np.power(result_df, 1 / len(values)).fillna(0)

    return result_df



import pandas as pd
import numpy as np
import anndata as ad
import pickle
from typing import Dict, List, Optional, Union
from copy import deepcopy
import json
from tqdm import tqdm

from . import tl


class CCIData:
    """
    Class to store and manage Cell-Cell Interaction (CCI) data
    
    Attributes:
        metadata (Dict): Metadata for sample
        n_spots (int): Number of spots in the sample
        cci_scores (Dict): CCI score dataframe for each LR pair
        p_values (Dict): P-values dataframe for each LR pair
        adata (AnnData): AnnData object
        networks (Dict): Calculated CCI networks
        other_metadata (Dict): Other metadata
        assays (Dict): Assays for the sample
    """
    
    def __init__(
        self, 
        cci_scores: Dict = None, 
        p_values: Dict = None,
        n_spots: int = None, 
        platform: str = None,
        network: pd.DataFrame = None,
        network_p_values: pd.DataFrame = None,
        adata: ad.AnnData = None,
        other_metadata: Dict = None,
        assays: Dict = None
        ):
        self.metadata = {}
        self.assays = {}
        self.adata = adata
        self.assays['raw'] = {}
        
        self.metadata['n_spots'] = n_spots
        self.metadata['platform'] = platform
        
        if other_metadata:
            self.metadata.update(other_metadata)

        if cci_scores is not None:
            self.assays['raw']['cci_scores'] = cci_scores

            if p_values is not None:
                self.assays['raw']['p_values'] = p_values
                
        if network is not None:
            self.assays['raw']['network'] = network
            
            if network_p_values is not None:
                self.assays['raw']['network_p_values'] = network_p_values
                
        if assays:
            self.assays = assays
                
    
    def __repr__(self):
        """Print assays (with number of LR pairs) and metadata"""
        
        assays = {}
        for assay in self.assays.keys():
            if 'cci_scores' in self.assays[assay].keys():
                assays[assay] = f"{len(self.assays[assay]['cci_scores'])} LR pairs"
            elif 'network' in self.assays[assay].keys():
                assays[assay] = "network"
            else:
                assays[assay] = "none"
            
        if self.adata is not None:
            return f"CCIData object with assays: {assays}, metadata: {self.metadata}, \
                and AnnData object"
        
        return f"CCIData object with assays: {assays} and metadata: {self.metadata}"
    
    
    def __str__(self):
        """Print assays (with number of LR pairs) and metadata"""
        
        return self.__repr__()
        
        
    def get_sample_metadata(self, sample_id: str) -> Dict:
        """
        Get metadata for a sample
        
        Args:
            sample_id: Sample ID
        
        Returns:
            Metadata for the sample
        """
        return self.metadata.get(sample_id)
    
    
    def get_sample_n_spots(self, sample_id: str) -> Optional[int]:
        """
        Get number of spots for a sample
        
        Args:
            sample_id: Sample ID
        
        Returns:
            Number of spots for the sample
        """
        return self.metadata['n_spots'].get(sample_id)
    
    
    def get_sample_cci_scores(self, sample_id: str) -> Optional[pd.DataFrame]:
        """
        Get CCI scores for a sample
        
        Args:
            sample_id: Sample ID
        
        Returns:
            CCI scores for the sample
        """
        return self.cci_scores.get(sample_id)
    
    
    def get_sample_p_values(self, sample_id: str) -> Optional[pd.DataFrame]:
        """
        Get p-values for a sample
        
        Args:
            sample_id: Sample ID
        
        Returns:
            P-values for the sample
        """
        return self.p_values.get(sample_id)
    
    
    def get_adata(self) -> Optional[ad.AnnData]:
        """
        Get AnnData object
        
        Returns:
            AnnData object
        """
        return self.adata
    
    
    def get_cell_types(self, assay: str = "raw") -> List[str]:
        """
        Get cell types in a sample
        
        Args:
            assay: Assay to get cell types from
        
        Returns:
            List of cell types in the sample
        """
        cell_types = []
        
        if 'cci_scores' in self.assays[assay].keys():
            for dfs in self.assays[assay]['cci_scores'].values():
                cell_types.extend(dfs.index)
        elif 'network' in self.assays[assay].keys():
            cell_types.extend(self.assays[assay]['network'].index)
        else:
            raise ValueError("No cell types found in sample.")
            
        return list(set(cell_types))
    
    
    def copy(self) -> 'CCIData':
        """
        Create a copy of the CCIData object
        
        Returns:
            Copy of the CCIData object
        """
        cci_data = CCIData(
            assays=deepcopy(self.assays),
            other_metadata=deepcopy(self.metadata),
            adata=deepcopy(self.adata)
        )
        
        return cci_data
    
    
    def rename_cell_types(self, 
                         replacements: Dict[str, str],
                         assay = None) -> 'CCIData':
        """Renames cell types in a CCIData.

        Args:
            replacements (dict): A dictionary of replacements, where the keys are the old cell type names and the values are the new cell type names.
            assay (str): The assay to rename the cell types in. If None, all assays are renamed.

        Returns:
            CCIData: A new CCIData object with the cell types renamed.
        """

        renamed_cci_data = self.copy()
        
        if assay is not None:
            for key in renamed_cci_data.assays[assay].keys():
                if key == 'network' or key == 'overall':
                    renamed_cci_data.assays[assay][key].rename(
                        index=replacements,
                        columns=replacements,
                        inplace=True)
                    
                else:
                    for lr_pair in renamed_cci_data.assays[assay][key].keys():
                        renamed_cci_data.assays[assay][key][lr_pair].rename(
                        index=replacements,
                        columns=replacements,
                        inplace=True)
            
        else:
            for assay in renamed_cci_data.assays.keys():
                for key in renamed_cci_data.assays[assay].keys():
                    if key == 'network' or key == 'overall':
                        renamed_cci_data.assays[assay][key].rename(
                            index=replacements,
                            columns=replacements,
                            inplace=True)
                        
                    else:
                        for lr_pair in renamed_cci_data.assays[assay][key].keys():
                            renamed_cci_data.assays[assay][key][lr_pair].rename(
                            index=replacements,
                            columns=replacements,
                            inplace=True)

        return renamed_cci_data
    
    
    def merge_cell_types(self,
                         cell_types: List[str],
                         new_cell_type: str,
                         assay = None) -> 'CCIData':
        """Merges cell types in a CCIData.
        
        Args:
            cell_types (list): A list of cell types to merge.
            new_cell_type (str): The name of the new cell type after merging.
            assay (str): The assay to merge the cell types in. If None, all assays are merged.
                
        Returns:
            CCIData: A new CCIData object with the cell types merged.
        """
        merged_cci_data = self.copy()
        assays = merged_cci_data.assays.keys()
        
        if assay is not None:
            assays = [assay]
            
        for assay in assays:
            for key in merged_cci_data.assays[assay].keys():
                if key == 'network' or key == 'overall':
                    df = merged_cci_data.assays[assay][key]
                    # Sum the rows
                    row_sums = df.loc[cell_types].sum()
                    # Sum the columns 
                    col_sums = df[cell_types].sum()
                    # Drop original cell types
                    df = df.drop(cell_types, axis=0)
                    df = df.drop(cell_types, axis=1) 
                    # Add new merged cell type
                    df.loc[new_cell_type] = row_sums
                    df[new_cell_type] = col_sums
                    merged_cci_data.assays[assay][key] = df
                elif key == 'cci_scores':
                    for lr_pair in merged_cci_data.assays[assay][key].keys():
                        df = merged_cci_data.assays[assay][key][lr_pair]
                        # Sum the rows
                        row_sums = df.loc[cell_types].sum()
                        # Sum the columns
                        col_sums = df[cell_types].sum()
                        # Drop original cell types
                        df = df.drop(cell_types, axis=0)
                        df = df.drop(cell_types, axis=1)
                        # Add new merged cell type
                        df.loc[new_cell_type] = row_sums
                        df[new_cell_type] = col_sums
                        merged_cci_data.assays[assay][key][lr_pair] = df
                elif key == 'p_values':
                    for lr_pair in merged_cci_data.assays[assay][key].keys():
                        df = merged_cci_data.assays[assay][key][lr_pair]
                        # Sum the rows
                        row_min = df.loc[cell_types].min()
                        # Sum the columns
                        col_min = df[cell_types].min()
                        # Drop original cell types
                        df = df.drop(cell_types, axis=0)
                        df = df.drop(cell_types, axis=1)
                        # Add new merged cell type
                        df.loc[new_cell_type] = row_min
                        df[new_cell_type] = col_min
        
        return merged_cci_data
    
    
    def subset_lrs(self,
                   lr_pairs: List[str],
                   assay = "raw",
                   new_assay = "subset") -> 'CCIData':
        """Subsets the LR pairs in a CCIData object.
    
        Args:
            lr_pairs (list): A list of LR pairs to include in the subsetted data.
            assay (str): The assay to subset the LR pairs in. If None, all assays are subsetted.
            new_assay (str): The name of the new assay after subsetting.
            
        Returns:
            CCIData: A new CCIData object with the LR pairs subsetted.
        """
        
        subsetted_cci_data = self.copy()
        subsetted_cci_data.assays[new_assay] = {}
        subsetted_cci_data.assays[new_assay]['cci_scores'] = \
            {lr: self.assays[assay]['cci_scores'][lr] for lr in lr_pairs}
        subsetted_cci_data.assays[new_assay]['p_values'] = \
            {lr: self.assays[assay]['p_values'][lr] for lr in lr_pairs}
        
        return subsetted_cci_data
    
    
    def scale(self, 
              scale_factor: float,
              assay = "raw",
              new_assay = "scaled") -> 'CCIData':
        """Scales the CCI scores in a CCIData object.

        Args:
            scale_factor (float): The factor to scale the CCI scores by.
            assay (str): The assay to scale the CCI scores in. If None, all assays are scaled.
            new_assay (str): The name of the new assay after scaling.

        Returns:
            CCIData: A new CCIData object with the CCI scores scaled.
        """

        scaled_cci_data = self.copy()
        scaled_cci_data.assays[new_assay] = self.assays[assay].copy()
        
        for lr_pair in self.assays[assay]['cci_scores'].keys():
            scaled_cci_data.assays[new_assay]['cci_scores'][lr_pair] = \
                self.assays[assay]['cci_scores'][lr_pair] * scale_factor
        
        return scaled_cci_data
    
    
    def scale_by_nspots(self, 
                        assay = "raw",
                        new_assay = "scaled") -> 'CCIData':
        """Scales the CCI scores in a CCIData object by the number of spots.

        Args:
            assay (str): The assay to scale the CCI scores in. If None, all assays are scaled.
            new_assay (str): The name of the new assay after scaling.

        Returns:
            CCIData: A new CCIData object with the CCI scores scaled.
        """

        if 'n_spots' not in self.metadata.keys():
            raise ValueError("'n_spots' not found in metadata.")
        
        scale_factor = 1e6 / self.metadata['n_spots']
        return self.scale(scale_factor, assay, new_assay)
    
    
    def filter_by_p_vals(self, 
                          cutoff: float = 0.05, 
                          assay = "raw",
                          new_assay = "filtered") -> 'CCIData':
        """Filters the CCI scores in a CCIData object by p-value.

        Args:
            cutoff (float): The p-value cutoff to filter the CCI scores by.
            assay (str): The assay to filter the CCI scores in. If None, all assays are filtered.
            new_assay (str): The name of the new assay after filtering.

        Returns:
            CCIData: A new CCIData object with the CCI scores filtered.
        """

        filtered_cci_data = self.copy()
        filtered_cci_data.assays[new_assay] = self.assays[assay].copy()
        
        for lr_pair, df in self.assays[assay]['cci_scores'].items():
                for i, row in df.iterrows():
                    for j in row.index:
                        p_vals = self.assays[assay]['p_values'][lr_pair]
                        if p_vals.loc[i, j] > cutoff:
                            df.loc[i, j] = 0
                filtered_cci_data.assays[new_assay]['cci_scores'][lr_pair] = df
                
        return filtered_cci_data
    
    
    def calc_overall(self, 
                     assay = "raw", 
                     name = "overall", 
                     normalize = True) -> 'CCIData':
        """Calculates the overall CCI scores in a CCIData object.

        Args:
            assay (str): The assay to calculate the overall CCI scores in. If None, all assays are calculated
            name (str): The name of the network after calculating the overall CCI scores.
            normalize (bool): If True, normalize each LR CCI score network. Defaults to True.
            
        Returns:
            CCIData: A new CCIData object with the overall CCI scores calculated.
        """
        
        overall_cci_data = self.copy()
        
        sample = overall_cci_data.assays[assay]['cci_scores']
        
        total = None
        for lr_pair in sample.keys():
            df_sum = sample[lr_pair].sum().sum()
            if df_sum > 0:
                if total is not None:
                    total, sample[lr_pair] = tl.align_dataframes(total, sample[lr_pair])
                    if normalize:
                        total = total + sample[lr_pair] / df_sum
                    else:
                        total = total + sample[lr_pair]
                    total = total.fillna(0)
                else:
                    if normalize:
                        total = sample[lr_pair] / df_sum
                    else:
                        total = sample[lr_pair]
                    total = total.fillna(0)

        total = total / total.sum().sum()
        total = total.fillna(0)

        overall_cci_data.assays[assay][name] = total
        
        return overall_cci_data
    
    
    def get_lr_proportions(
        self,
        sender: str = None,
        reciever: str = None,
        assay: str = "raw",
        key: str = "cci_scores"
        ) -> dict:
        """Calculates the proportion of each LR pair in a sample for a specific cell type sender and receiver.

        Args:
            sender (str): The sender cell type.
            reciever (str): The receiver cell type.
            assay (str) (optional): The assay to use. Defaults to 'raw'.
            key (str) (optional): The key to use. Defaults to 'cci_scores'.

        Returns:
            dict: A list of LR pairs and proportion of its weighting.
        """

        if assay not in self.assays:
            raise ValueError(f"Assay {assay} not found in sample.")
        
        if key not in self.assays[assay]:
            raise ValueError(f"Key {key} not found in sample.")
        
        if sender is None and reciever is None:
            raise ValueError("Please provide a sender or receiver cell type.")
        
        lr_pairs = self.assays[assay][key].keys()
        for lr in lr_pairs:
            if type(self.assays[assay][key][lr]) != pd.DataFrame:
                del self.assays[assay][key][lr]

        if sender is not None:
            subset = {
                lr: df.loc[[sender]] for lr, df in self.assays[assay][key].items()
                if sender in df.index
                }
            
        if reciever is not None:
            subset = {
                lr: df[[reciever]] for lr, df in self.assays[assay][key].items()
                if reciever in df.columns
                }
            
        subset = {
            key: df
            for key, df in subset.items()
            if not df.map(lambda x: x == 0).all().all()
        }

        subset = {
            key: df
            for key, df in subset.items()
            if not df.map(lambda x: x == 0).all().all()
        }

        lr_props = {}
        total = 0
        
        if sender is not None and reciever is not None:
            for lr_pair in set(subset.keys()):
                score = subset[lr_pair].at[sender, reciever]
                total += score

            for lr_pair in set(subset.keys()):
                lr_props[lr_pair] = subset[lr_pair].at[sender, reciever] / total
                
        if sender is None:
            # sum across all sender cell types
            for lr_pair in set(subset.keys()):
                score = subset[lr_pair].loc[:, reciever].sum()
                total += score
                
            for lr_pair in set(subset.keys()):
                lr_props[lr_pair] = subset[lr_pair].loc[:, reciever].sum() / total
                
        if reciever is None:
            # sum across all receiver cell types
            for lr_pair in set(subset.keys()):
                score = subset[lr_pair].loc[sender, :].sum()
                total += score
                
            for lr_pair in set(subset.keys()):
                lr_props[lr_pair] = subset[lr_pair].loc[sender, :].sum() / total

        lr_props = dict(sorted(lr_props.items(), key=lambda item: item[1], 
                               reverse=True))

        return lr_props


    def get_p_vals(
        self,
        assay: str = "raw",
        key: str = "p_values",
        sender: str = None,
        reciever: str = None
        ) -> dict:
        """Returns the p-values for each LR pair in a sample for a specific sender and receiver cell type.
        
        Args:
            assay (str) (optional): The assay to use. Defaults to 'raw'.
            key (str) (optional): The key to use. Defaults to 'p_values'.
            sender (str) (optional): The sender cell type. Defaults to None.
            reciever (str) (optional): The receiver cell type. Defaults to None.
            
        Returns:
            dict: A dictionary of LR pairs and their p-values.
        """
        
        if assay not in self.assays:
            raise ValueError(f"Assay {assay} not found in sample.")
        
        if key not in self.assays[assay]:
            raise ValueError(f"Key {key} not found in sample.")
        
        if sender is None or reciever is None:
            raise ValueError("Please provide a sender and receiver cell type.")
        
        lr_pairs = self.assays[assay][key].keys()
        for lr in lr_pairs:
            if type(self.assays[assay][key][lr]) != pd.DataFrame:
                del self.assays[assay][key][lr]
                
        subset = {
            lr: df.loc[[sender], [reciever]] for lr, df in 
            self.assays[assay][key].items()
            if sender in df.index and reciever in df.columns
            }
        
        subset = {
            key: df
            for key, df in subset.items()
            if not df.map(lambda x: x == 0).all().all()
        }
        
        subset = {
            key: df
            for key, df in subset.items()
            if not df.map(lambda x: x == 0).all().all()
        }
        
        p_vals = {}
        for lr_pair in set(subset.keys()):
            p_vals[lr_pair] = subset[lr_pair].at[sender, reciever]
            
        return p_vals


    def save(self, path: str):
        """Saves CCIData object to JSON or pkl file.
        
        Args:
            path (str): Path to save the JSON or pkl file
        """
        if not path.endswith('.json'):
            path += '.json'
            
        if path.endswith('.json'):
            with open(path, 'w') as f:
                json.dump(self.to_dict(), f)
                
        if path.endswith('.pkl'):
            with open(path, 'wb') as f:
                pickle.dump(self, f)

            
    def to_dict(self) -> dict:
        """Convert CCIData object to a JSON-serializable dictionary.
        
        Returns:
            dict: Dictionary representation of the CCIData object
        """
        data_dict = {
            'metadata': self.metadata,
            'assays': {}
        }
        
        # Convert assays
        for assay_name, assay in self.assays.items():
            data_dict['assays'][assay_name] = {}
            for key, value in assay.items():
                if key in ['cci_scores', 'p_values']:
                    # Convert dict of DataFrames
                    data_dict['assays'][assay_name][key] = {
                        k: v.to_dict() for k,v in value.items()
                    }
                elif isinstance(value, pd.DataFrame):
                    # Convert single DataFrame
                    data_dict['assays'][assay_name][key] = value.to_dict()
                else:
                    data_dict['assays'][assay_name][key] = value
                    
        return data_dict


    def create_pathway_assay(
        self,
        assay: str = "raw",
        gsea_results: pd.DataFrame = None,
        strict: bool = True,
        cutoff: float = 0.05,
        assay_name = "pathway"
        ) -> 'CCIData':
        """Creates a pathway assay from GSEA results.
        
        Args:
            assay (str): The assay to use. Defaults to 'raw'.
            gsea_results (pd.DataFrame): GSEA results. Defaults to None.
            strict (bool): If True, both ligand and receptor must be in the gene list. If False, only one must be in the gene list. Defaults to True.
            cutoff (float): The p-value cutoff to filter the GSEA results by. Defaults to 0.05.
            assay_name (str): The name of the new assay. Defaults to 'pathway'.
            
        Returns:
            CCIData: A new CCIData object with the pathway assay.
        """
        
        copy = self.copy()
        grouped_cci_scores = {}

        with tqdm(total=len(gsea_results), desc="Converting to pathways") as pbar:
            for term in gsea_results['Term']:
                
                filtered_df = gsea_results[gsea_results['Term'] == term]
                if filtered_df['Adjusted P-value'].values[0] > cutoff:
                    tqdm.update(pbar, 1)
                    continue
                
                gene_list = filtered_df['Genes'].tolist()
                genes = []
                for gene in gene_list:
                    genes.extend(gene.lower().split(";"))
                    
                cci_scores = []
                        
                for key in self.assays[assay]['cci_scores'].keys():
                    lig, rec = key.lower().split("_")
                    if strict:
                        if lig in genes and rec in genes:
                            cci_scores.append(self.assays[assay]['cci_scores'][key])
                    else:
                        if lig in genes or rec in genes:
                            cci_scores.append(self.assays[assay]['cci_scores'][key])
                            
                total = None      
                for df in cci_scores:
                    if df.sum().sum() > 0:
                        if total is not None:
                            total = total + df
                            total = total.fillna(0)
                        else:
                            total = df
                            total = total.fillna(0)
                            
                if total is None:
                    tqdm.update(pbar, 1)
                    continue
                
                total = total / total.sum().sum()
                total = total.fillna(0)

                grouped_cci_scores[term] = total
                tqdm.update(pbar, 1)
                        
        copy.assays[assay_name] = {'cci_scores': grouped_cci_scores}
        
        return copy



from matplotlib import pyplot as plt
import numpy as np
from matplotlib.patches import Arc, Wedge
from matplotlib.path import Path
import matplotlib.patches as patches
import matplotlib.colors as plt_colors
import matplotlib.cm as cm


def hex2rgb(c):
    return tuple(int(c[i: i + 2], 16) / 256.0 for i in (1, 3, 5))


def polar2xy(r, theta):
    return np.array([r * np.cos(theta), r * np.sin(theta)])


def ChordArc(
    start1=0,
    end1=60,
    start2=180,
    end2=240,
    radius=1.0,
    chordwidth=0.7,
    ax=None,
    color=(1, 0, 0),
):
    # start, end should be in [0, 360)
    if start1 > end1:
        start1, end1 = end1, start1
    if start2 > end2:
        start2, end2 = end2, start2
    start1 *= np.pi / 180.0
    end1 *= np.pi / 180.0
    start2 *= np.pi / 180.0
    end2 *= np.pi / 180.0
    opt1 = 4.0 / 3.0 * np.tan((end1 - start1) / 4.0) * radius
    opt2 = 4.0 / 3.0 * np.tan((end2 - start2) / 4.0) * radius
    rchord = radius * (1 - chordwidth)
    verts = [
        polar2xy(radius, start1),
        polar2xy(radius, start1) + polar2xy(opt1, start1 + 0.5 * np.pi),
        polar2xy(radius, end1) + polar2xy(opt1, end1 - 0.5 * np.pi),
        polar2xy(radius, end1),
        polar2xy(rchord, end1),
        polar2xy(rchord, start2),
        polar2xy(radius, start2),
        polar2xy(radius, start2) + polar2xy(opt2, start2 + 0.5 * np.pi),
        polar2xy(radius, end2) + polar2xy(opt2, end2 - 0.5 * np.pi),
        polar2xy(radius, end2),
        polar2xy(rchord, end2),
        polar2xy(rchord, start1),
        polar2xy(radius, start1),
    ]

    codes = [
        Path.MOVETO,
        Path.CURVE4,
        Path.CURVE4,
        Path.CURVE4,
        Path.CURVE4,
        Path.CURVE4,
        Path.CURVE4,
        Path.CURVE4,
        Path.CURVE4,
        Path.CURVE4,
        Path.CURVE4,
        Path.CURVE4,
        Path.CURVE4,
    ]

    if ax is None:
        return verts, codes
    else:
        path = Path(verts, codes)
        patch = patches.PathPatch(
            path, facecolor=color + (0.5,), edgecolor=color + (0.4,), lw=0.3
        )
        ax.add_patch(patch)


def selfChordArc(start=0, end=60, radius=1.0, chordwidth=0.7, ax=None, color=(1, 0, 0)):
    # start, end should be in [0, 360)
    if start > end:
        start, end = end, start
    start *= np.pi / 180.0
    end *= np.pi / 180.0
    opt = 4.0 / 3.0 * np.tan((end - start) / 4.0) * radius
    rchord = radius * (1 - chordwidth)
    verts = [
        polar2xy(radius, start),
        polar2xy(radius, start) + polar2xy(opt, start + 0.5 * np.pi),
        polar2xy(radius, end) + polar2xy(opt, end - 0.5 * np.pi),
        polar2xy(radius, end),
        polar2xy(rchord, end),
        polar2xy(rchord, start),
        polar2xy(radius, start),
    ]

    codes = [
        Path.MOVETO,
        Path.CURVE4,
        Path.CURVE4,
        Path.CURVE4,
        Path.CURVE4,
        Path.CURVE4,
        Path.CURVE4,
    ]

    if ax is None:
        return verts, codes
    else:
        path = Path(verts, codes)
        patch = patches.PathPatch(
            path, facecolor=color + (0.5,), edgecolor=color + (0.4,), lw=0.3
        )
        ax.add_patch(patch)


def chordDiagram(X, ax, colors=None, width=0.1, pad=2, chordwidth=0.7, lim=1.1):
    """Plot a chord diagram
    Parameters
    ----------
    X :
        flux data, X[i, j] is the flux from i to j
    ax :
        matplotlib `axes` to show the plot
    colors : optional
        user defined colors in rgb format. Use function hex2rgb() to convert hex color to rgb color. Default: d3.js category10
    width : optional
        width/thickness of the ideogram arc
    pad : optional
        gap pad between two neighboring ideogram arcs, unit: degree, default: 2 degree
    chordwidth : optional
        position of the control points for the chords, controlling the shape of the chords
    """
    # X[i, j]:  i -> j
    x = X.sum(axis=1)  # sum over rows
    ax.set_xlim(-lim, lim)
    ax.set_ylim(-lim, lim)
    diam = 1.8

    if colors is None:
        # use d3.js category10
        # https://github.com/d3/d3-3.x-api-reference/blob/master/Ordinal-Scales.md#category10
        colors = [
            "#1f77b4",
            "#ff7f0e",
            "#2ca02c",
            "#d62728",
            "#9467bd",
            "#8c564b",
            "#e377c2",
            "#7f7f7f",
            "#bcbd22",
            "#17becf",
        ]
        if len(x) > 10:
            print("x is too large! Use x smaller than 10")
    if isinstance(colors[0], str):
        colors = [hex2rgb(colors[i]) for i in range(len(x))]

    # find position for each start and end
    y = x / np.sum(x).astype(float) * (360 - pad * len(x))

    pos = {}
    arc = []
    nodePos = []
    start = 0
    for i in range(len(x)):
        end = start + y[i]
        arc.append((start, end))
        angle = 0.5 * (start + end)
        # print(start, end, angle)
        if -30 <= angle <= 210:
            angle -= 90
        else:
            angle -= 270
        nodePos.append(
            tuple(
                polar2xy((diam / 2) + diam * 0.05, 0.5 * (start + end) * np.pi / 180.0)
            )
            + (angle,)
        )

        x[x == 0] = 1
        z = (X[i, :] / x[i].astype(float)) * (end - start)
        z[np.isnan(z)] = 0

        ids = np.argsort(z)
        z0 = start
        for j in ids:
            pos[(i, j)] = (z0, z0 + z[j])
            z0 += z[j]
        start = end + pad

    for i in range(len(x)):
        start, end = arc[i]
        # This draws the outter ring #
        # IdeogramArc(start=start, end=end, radius=1.0, ax=ax,
        #            color=colors[i], width=width)
        a = Arc(
            (0, 0),
            diam,
            diam,
            angle=0,
            theta1=start,
            theta2=end,
            color=colors[i],
            lw=10,
        )
        ax.add_patch(a)
        start, end = pos[(i, i)]
        # This draws the paths to itself #
        if end - start < 180:  # Indicates this method will work fine !
            selfChordArc(
                start,
                end,
                radius=1.0 - width,
                color=colors[i],
                chordwidth=chordwidth * 0.7,
                ax=ax,
            )
        else:  # Need to use a wedge because the arch distorts past 180-degrees
            path = Wedge(0, diam / 2, start, end, color=colors[i] + (0.5,))
            ax.add_patch(path)
        for j in range(i):
            if X[i, j] == 0 and X[j, i] == 0:  # don't draw anything for no interaction
                continue
            color = colors[i]
            if X[i, j] > X[j, i]:  # Color by the dominant signal #
                color = colors[j]
            start1, end1 = pos[(i, j)]
            start2, end2 = pos[(j, i)]
            ChordArc(
                start1,
                end1,
                start2,
                end2,
                radius=1.0 - width,
                color=color,
                chordwidth=chordwidth,
                ax=ax,
            )

    return nodePos
